from __future__ import annotations

import json as jsonlib
from typing import TYPE_CHECKING, TextIO, TypeAlias

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import Plugin, arg, export, internal

if TYPE_CHECKING:
    from dissect.target.target import Target

_EsxConf: TypeAlias = dict[str, "str | int | bool | _EsxConf"]


class EsxConfPlugin(Plugin):
    """Plugin to interact with ``esxconf`` configuration."""

    __namespace__ = "esxconf"

    def __init__(self, target: Target):
        super().__init__(target)
        self._config = None

        if (path := target.fs.path("/etc/vmware/esx.conf")).exists():
            with path.open("rt") as fh:
                self._config = parse_esx_conf(fh)

    def check_compatible(self) -> None:
        if self.target.os != "esxi":
            raise UnsupportedPluginError("ESXi specific plugin loaded on non-ESXi target")

        if not self._config:
            raise UnsupportedPluginError("esx.conf not found on target")

    def _cfg(self, path: str) -> str | int | bool | _EsxConf | None:
        if not self._config:
            self.target.log.warning("No ESXi config!")
            return None

        value_name = path.strip("/").split("/")[-1]
        obj = _traverse(path, self._config)

        if not value_name and obj:
            return obj

        return obj.get(value_name) if obj else None

    @internal
    def get(self, path: str) -> str | int | bool | _EsxConf | None:
        """Get esxconf value at the specified path."""
        return self._cfg(path)

    @export(output="none")
    @arg("path", help="config path")
    @arg("-j", "--json", action="store_true", help="output in JSON format")
    def __call__(self, path: str, json: bool) -> None:
        """Dump esxconf value at the specified path."""
        obj = self._cfg(path)

        if json:
            print(jsonlib.dumps(obj, indent=4, sort_keys=True))
        else:
            print(obj)


def parse_esx_conf(fh: TextIO) -> _EsxConf:
    config = {}
    for line in fh:
        if not (line := line.strip()):
            continue

        key, _, value = line.partition("=")
        key = key.strip().strip("/")
        value = value.strip().strip('"')

        if value == "true":
            value = True
        elif value == "false":
            value = False
        elif value.isnumeric():
            value = int(value)

        value_name = key.split("/")[-1]
        obj = _traverse(key, config, create=True)
        obj[value_name] = value

    return config


def _traverse(path: str, obj: _EsxConf, create: bool = False) -> _EsxConf | None:
    parts = path.strip("/").split("/")
    path_parts = parts[:-1]
    for part in path_parts:
        array_idx = None
        if part.endswith("]"):
            part, _, rest = part.partition("[")
            array_idx = rest.strip("]")

        if part not in obj:
            if create:
                obj[part] = {}
            else:
                return None

        obj = obj[part]
        if array_idx:
            if array_idx not in obj:
                if create:
                    obj[array_idx] = {}
                else:
                    return None
            obj = obj[array_idx]

    return obj



from __future__ import annotations

import json
import re
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

from dissect.util.ts import from_unix

from dissect.target.exceptions import FileNotFoundError, UnsupportedPluginError
from dissect.target.helpers.fsutil import basename, open_decompress
from dissect.target.plugin import export
from dissect.target.plugins.apps.webserver.webserver import (
    WebserverAccessLogRecord,
    WebserverPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

LOG_FILE_REGEX = re.compile(r"(log|output file) (?P<log_file>.*)( \{)?$")
LOG_REGEX = re.compile(
    r'(?P<remote_ip>.*?) - - \[(?P<ts>\d{2}\/[A-Za-z]{3}\/\d{4}:\d{2}:\d{2}:\d{2} (\+|\-)\d{4})\] "(?P<method>.*?) (?P<uri>.*?) ?(?P<protocol>HTTP\/.*?)?" (?P<status_code>\d{3}) (?P<bytes_sent>-|\d+)'  # noqa: E501
)


class CaddyPlugin(WebserverPlugin):
    """Caddy webserver plugin."""

    __namespace__ = "caddy"

    DEFAULT_CONFIG_PATH = "/etc/caddy/Caddyfile"

    def __init__(self, target: Target):
        super().__init__(target)
        self.log_paths = self.get_log_paths()

    def check_compatible(self) -> None:
        if not len(self.log_paths):
            raise UnsupportedPluginError("No Caddy paths found")

    def get_log_paths(self) -> list[Path]:
        log_paths = []

        # Add any well known default Caddy log locations
        log_paths.extend(self.target.fs.path("/var/log").glob("caddy_access.log*"))

        # Check for custom paths in Caddy config
        if (config_file := self.target.fs.path(self.DEFAULT_CONFIG_PATH)).exists():
            found_roots = []
            for line in config_file.open("rt"):
                line = line.strip()
                if not line:
                    continue

                if line.startswith("#"):
                    line = line[1:].strip()

                if "root " in line:
                    found_roots.append(line.split("root ")[1].strip())

                if "log " in line or "output file " in line:
                    if line == "log {":
                        continue

                    match = LOG_FILE_REGEX.match(line)
                    if not match:
                        self.target.log.warning("Could not determine log path in %s: %s", config_file, line)
                        continue

                    match = match.groupdict()
                    log_path = match["log_file"].replace(" {", "").split(" ")[-1]

                    parent_folders = []
                    # Search all root folders we found earlier with the current relative log path we found.
                    if log_path.startswith("/"):
                        parent_folders.append(self.target.fs.path(log_path).parent)
                    else:
                        if len(found_roots) == 0:
                            self.target.log.warning(
                                "Can not infer absolute log path from relative path: Caddyfile root is not configured."
                            )
                            continue

                        for root in found_roots:
                            # Logs will be located one folder higher than the defined root.
                            root_parent = self.target.fs.path(root).parent
                            parent_folders.append(root_parent.joinpath(log_path).parent)

                    for parent_folder in parent_folders:
                        log_paths.extend(
                            path for path in parent_folder.glob(f"{basename(log_path)}*") if path not in log_paths
                        )

        return log_paths

    def _get_paths(self) -> Iterator[Path]:
        yield from self.log_paths

    def _get_auxiliary_paths(self) -> Iterator[Path]:
        yield from {Path(self.DEFAULT_CONFIG_PATH)}

    @export(record=WebserverAccessLogRecord)
    def access(self) -> Iterator[WebserverAccessLogRecord]:
        """Parses Caddy V1 CRF and Caddy V2 JSON access logs.

        References:
            - https://caddyserver.com/docs/caddyfile/directives/log#format-modules
        """
        for path in self.log_paths:
            try:
                path = path.resolve(strict=True)
                for line in open_decompress(path, "rt"):
                    line = line.strip()
                    if not line:
                        continue

                    # Parse a JSON log line
                    if line.startswith('{"'):
                        try:
                            log = json.loads(line)
                        except json.decoder.JSONDecodeError:
                            self.target.log.warning("Could not decode Caddy JSON log line: %s (%s)", line, path)
                            continue

                        yield WebserverAccessLogRecord(
                            ts=from_unix(log["ts"]),
                            remote_ip=log["request"]["remote_ip"],
                            method=log["request"]["method"],
                            uri=log["request"]["uri"],
                            protocol=log["request"]["proto"],
                            status_code=log["status"],
                            bytes_sent=log["size"],
                            source=path,
                            _target=self.target,
                        )

                    # Try to parse a CLF log line
                    else:
                        match = LOG_REGEX.match(line)
                        if not match:
                            self.target.log.warning(
                                "Could not match Caddy regex format for log line: %s (%s)", line, path
                            )
                            continue

                        log = match.groupdict()
                        yield WebserverAccessLogRecord(
                            ts=datetime.strptime(log["ts"], "%d/%b/%Y:%H:%M:%S %z"),
                            webserver="caddy",
                            remote_ip=log["remote_ip"],
                            method=log["method"],
                            uri=log["uri"],
                            protocol=log["protocol"],
                            status_code=log["status_code"],
                            bytes_sent=log["bytes_sent"].strip("-") or 0,
                            source=path,
                            _target=self.target,
                        )
            except FileNotFoundError:
                self.target.log.warning("Caddy log file configured but could not be found (dead symlink?): %s", path)
            except Exception as e:
                self.target.log.warning("An error occured parsing Caddy log file %s: %s", path, str(e))
                self.target.log.debug("", exc_info=e)



"""A pathlib.Path compatible implementation for dissect.target.

This allows for the majority of the pathlib.Path API to "just work" on dissect.target filesystems.

Most of this consists of subclassed internal classes with dissect.target specific patches,
but sometimes the change to a function is small, so the entire internal function is copied
and only a small part changed. To ease updating this code, the order of functions, comments
and code style is kept largely the same as the original pathlib.py.

Yes, we know, this is playing with fire and it can break on new CPython releases.

The implementation is split up in multiple files, one for each CPython version.
You're currently looking at the CPython 3.13 implementation.

Commit hash we're in sync with: 094d95f

Notes:
    - https://docs.python.org/3.13/whatsnew/3.13.html#pathlib
    - https://github.com/python/cpython/blob/3.13/Lib/pathlib/_local.py
"""

from __future__ import annotations

import posixpath
import sys
from glob import _Globber
from pathlib import Path, PurePath
from pathlib._abc import PathBase, UnsupportedOperation
from typing import IO, TYPE_CHECKING, ClassVar

from dissect.target import filesystem
from dissect.target.exceptions import FilesystemError, SymlinkRecursionError
from dissect.target.helpers import polypath
from dissect.target.helpers.compat import path_common

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem, FilesystemEntry
    from dissect.target.helpers.fsutil import stat_result


class _DissectParser:
    sep = "/"
    altsep = ""
    case_sensitive = False

    __variant_instances: ClassVar[dict[tuple[bool, str], _DissectParser]] = {}

    def __new__(cls, case_sensitive: bool = False, alt_separator: str = ""):
        idx = (case_sensitive, alt_separator)
        instance = cls.__variant_instances.get(idx, None)
        if instance is None:
            instance = object.__new__(cls)
            cls.__variant_instances[idx] = instance

        return instance

    def __init__(self, case_sensitive: bool = False, alt_separator: str = ""):
        self.altsep = alt_separator
        self.case_sensitive = case_sensitive

    def normcase(self, s: str) -> str:
        return s if self.case_sensitive else s.lower()

    def split(self, part: str) -> tuple[str, str]:
        return polypath.split(part, alt_separator=self.altsep)

    splitdrive = staticmethod(posixpath.splitdrive)

    def splitroot(self, part: str) -> tuple[str, str, str]:
        return polypath.splitroot(part, alt_separator=self.altsep)

    def join(self, *args) -> str:
        return polypath.join(*args, alt_separator=self.altsep)

    isjunction = staticmethod(path_common.isjunction)

    def isabs(self, path: str) -> bool:
        return polypath.isabs(str(path), alt_separator=self.altsep)

    realpath = staticmethod(path_common.realpath)


class _DissectGlobber(_Globber):
    @staticmethod
    def add_slash(path: TargetPath) -> TargetPath:
        return _GlobberTargetPath(path._fs, path, "")


class PureDissectPath(PurePath):
    _fs: Filesystem
    parser: _DissectParser = _DissectParser(case_sensitive=False)
    _globber = _DissectGlobber

    def __reduce__(self) -> tuple:
        raise TypeError("TargetPath pickling is currently not supported")

    def __init__(self, fs: Filesystem, *pathsegments):
        if not isinstance(fs, filesystem.Filesystem):
            raise TypeError(
                "invalid PureDissectPath initialization: missing filesystem, "
                "got {!r} (this might be a bug, please report)".format(fs, *pathsegments)
            )

        alt_separator = fs.alt_separator
        path_args = []
        for arg in pathsegments:
            if isinstance(arg, str):
                arg = polypath.normalize(arg, alt_separator=alt_separator)
            path_args.append(arg)

        super().__init__(*path_args)
        self._fs = fs
        self.parser = _DissectParser(alt_separator=fs.alt_separator, case_sensitive=fs.case_sensitive)

    def with_segments(self, *pathsegments) -> Self:
        return type(self)(self._fs, *pathsegments)

    # NOTE: This is copied from pathlib/_local.py
    # but turned into an instance method so we get access to the correct flavour
    def _parse_path(self, path: str) -> tuple[str, str, list[str]]:
        if not path:
            return "", "", []
        sep = self.parser.sep
        altsep = self.parser.altsep
        if altsep:
            path = path.replace(altsep, sep)
        drv, root, rel = self.parser.splitroot(path)
        if not root and drv.startswith(sep) and not drv.endswith(sep):
            drv_parts = drv.split(sep)
            if len(drv_parts) == 4 and drv_parts[2] not in "?.":
                # e.g. //server/share
                root = sep
            elif len(drv_parts) == 6:
                # e.g. //?/unc/server/share
                root = sep
        parsed = [sys.intern(str(x)) for x in rel.split(sep) if x and x != "."]
        return drv, root, parsed


class TargetPath(Path, PureDissectPath):
    __slots__ = ("_direntry", "_entry")

    @classmethod
    def _unsupported_msg(cls, attribute: str) -> str:
        return f"{cls.__name__}.{attribute} is unsupported"

    def get(self) -> FilesystemEntry:
        """Return the :class:`FilesystemEntry` for this path."""
        if not hasattr(self, "_entry"):
            self._entry = self._direntry.get() if hasattr(self, "_direntry") else self._fs.get(str(self))
        return self._entry

    def stat(self, *, follow_symlinks: bool = True) -> stat_result:
        """
        Return the result of the stat() system call on this path, like
        os.stat() does.
        """
        if follow_symlinks:
            return self.get().stat()
        return self.get().lstat()

    def exists(self, *, follow_symlinks: bool = True) -> bool:
        """
        Whether this path exists.

        This method normally follows symlinks; to check whether a symlink exists,
        add the argument follow_symlinks=False.
        """
        try:
            # .exists() must resolve possible symlinks
            self.stat(follow_symlinks=follow_symlinks)
        except (FilesystemError, ValueError):
            return False
        else:
            return True

    is_mount = PathBase.is_mount

    def is_junction(self) -> bool:
        """
        Whether this path is a junction.
        """
        return self.parser.isjunction(self)

    def open(
        self,
        mode: str = "rb",
        buffering: int = 0,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> IO:
        """Open file and return a stream.

        Supports a subset of features of the real pathlib.open/io.open.

        Note: in contrast to regular Python, the mode is binary by default. Text mode
        has to be explicitly specified. Buffering is also disabled by default.
        """
        return path_common.io_open(self, mode, buffering, encoding, errors, newline)

    def write_bytes(self, data: bytes) -> int:
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        raise UnsupportedOperation(self._unsupported_msg("write_bytes()"))

    def write_text(
        self, data: str, encoding: str | None = None, errors: str | None = None, newline: str | None = None
    ) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        raise UnsupportedOperation(self._unsupported_msg("write_text()"))

    def iterdir(self) -> Iterator[Self]:
        """Yield path objects of the directory contents.

        The children are yielded in arbitrary order, and the
        special entries '.' and '..' are not included.
        """
        with path_common.scandir(self) as scandir_it:
            for entry in scandir_it:
                name = entry.name
                child_path = self.joinpath(name)
                child_path._direntry = entry
                yield child_path

    def glob(
        self, pattern: str, *, case_sensitive: bool | None = None, recurse_symlinks: bool = False
    ) -> Iterator[Self]:
        """Iterate over this subtree and yield all existing files (of any
        kind, including directories) matching the given relative pattern.
        """
        return PathBase.glob(self, pattern, case_sensitive=case_sensitive, recurse_symlinks=recurse_symlinks)

    def rglob(
        self, pattern: str, *, case_sensitive: bool | None = None, recurse_symlinks: str = False
    ) -> Iterator[Self]:
        """Recursively yield all existing files (of any kind, including
        directories) matching the given relative pattern, anywhere in
        this subtree.
        """
        return PathBase.rglob(self, pattern, case_sensitive=case_sensitive, recurse_symlinks=recurse_symlinks)

    def walk(
        self, top_down: bool = True, on_error: Callable[[Exception], None] | None = None, follow_symlinks: bool = False
    ) -> Iterator[tuple[Self, list[str], list[str]]]:
        """Walk the directory tree from this directory, similar to os.walk()."""
        return PathBase.walk(self, top_down=top_down, on_error=on_error, follow_symlinks=follow_symlinks)

    def absolute(self) -> Self:
        """Return an absolute version of this path
        No normalization or symlink resolution is performed.

        Use resolve() to resolve symlinks and remove '..' segments.
        """
        raise UnsupportedOperation(self._unsupported_msg("absolute()"))

    @classmethod
    def cwd(cls) -> Self:
        """Return a new path pointing to the current working directory."""
        raise UnsupportedOperation(cls._unsupported_msg("cwd()"))

    def expanduser(self) -> Self:
        """Return a new path with expanded ~ and ~user constructs
        (as returned by os.path.expanduser)
        """
        raise UnsupportedOperation(self._unsupported_msg("expanduser()"))

    @classmethod
    def home(cls) -> Self:
        """Return a new path pointing to the user's home directory (as
        returned by os.path.expanduser('~')).
        """
        raise UnsupportedOperation(cls._unsupported_msg("home()"))

    def readlink(self) -> Self:
        """
        Return the path to which the symbolic link points.
        """
        return self.with_segments(self.get().readlink())

    # NOTE: We changed some of the error handling here to deal with our own exception types
    def resolve(self, strict: bool = False) -> Self:
        """
        Make the path absolute, resolving all symlinks on the way and also
        normalizing it.
        """

        s = self.parser.realpath(self, strict=strict)
        p = self.with_segments(s)

        # In non-strict mode, realpath() doesn't raise on symlink loops.
        # Ensure we get an exception by calling stat()
        if not strict:
            try:
                p.stat()
            except FilesystemError as e:
                if isinstance(e, SymlinkRecursionError):
                    raise
        return p

    def symlink_to(self, target: str, target_is_directory: bool = False) -> None:
        """
        Make this path a symlink pointing to the target path.
        Note the order of arguments (link, target) is the reverse of os.symlink.
        """
        raise UnsupportedOperation(self._unsupported_msg("symlink_to()"))

    def hardlink_to(self, target: str) -> None:
        """
        Make this path a hard link pointing to the same file as *target*.

        Note the order of arguments (self, target) is the reverse of os.link's.
        """
        raise UnsupportedOperation(self._unsupported_msg("hardlink_to()"))

    def touch(self, mode: int = 0o666, exist_ok: bool = True) -> None:
        """
        Create this file with the given access mode, if it doesn't exist.
        """
        raise UnsupportedOperation(self._unsupported_msg("touch()"))

    def mkdir(self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False) -> None:
        """
        Create a new directory at this given path.
        """
        raise UnsupportedOperation(self._unsupported_msg("mkdir()"))

    def rename(self, target: str) -> Self:
        """
        Rename this path to the target path.

        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.

        Returns the new Path instance pointing to the target path.
        """
        raise UnsupportedOperation(self._unsupported_msg("rename()"))

    def replace(self, target: str) -> Self:
        """
        Rename this path to the target path, overwriting if that path exists.

        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.

        Returns the new Path instance pointing to the target path.
        """
        raise UnsupportedOperation(self._unsupported_msg("replace()"))

    def chmod(self, mode: int, *, follow_symlinks: bool = True) -> None:
        """
        Change the permissions of the path, like os.chmod().
        """
        raise UnsupportedOperation(self._unsupported_msg("chmod()"))

    def lchmod(self, mode: int) -> None:
        """
        Like chmod(), except if the path points to a symlink, the symlink's
        permissions are changed, rather than its target's.
        """
        raise UnsupportedOperation(self._unsupported_msg("lchmod()"))

    def unlink(self, missing_ok: bool = False) -> None:
        """
        Remove this file or link.
        If the path is a directory, use rmdir() instead.
        """
        raise UnsupportedOperation(self._unsupported_msg("unlink()"))

    def rmdir(self) -> None:
        """
        Remove this directory.  The directory must be empty.
        """
        raise UnsupportedOperation(self._unsupported_msg("rmdir()"))

    def owner(self) -> str:
        """
        Return the login name of the file owner.
        """
        raise UnsupportedOperation(self._unsupported_msg("owner()"))

    def group(self) -> str:
        """
        Return the group name of the file gid.
        """
        raise UnsupportedOperation(self._unsupported_msg("group()"))


class _GlobberTargetPath(TargetPath):
    def __str__(self) -> str:
        # This is necessary because the _Globber class expects an added `/` at the end
        # However, only PurePathBase properly adds that, PurePath doesn't
        # We do want to operate on Path objects rather than strings, so do a little hack here
        return self._raw_path



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.regf.clsid import CLSIDPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_clsid_plugin(target_win_users: Target, hive_hklm: VirtualHive) -> None:
    clsid_key_name = "SOFTWARE\\Classes\\CLSID"
    clsid_key = VirtualKey(hive_hklm, clsid_key_name)

    subkey_name = "{20894375-46AE-46E2-BAFD-CB38975CDCE6}"
    subkey = VirtualKey(hive_hklm, subkey_name)
    subkey.add_value("(Default)", VirtualValue(hive_hklm, "(Default)", "ShareHandler Class"))

    value_str = "C:\\Users\\John\\AppData\\Local\\Microsoft\\OneDrive\\21.230.1107.0004\\FileSyncShell64.dll"
    subsubkey_name = "InprocServer32"
    subsubkey = VirtualKey(hive_hklm, subsubkey_name)
    subsubkey.add_value("(Default)", VirtualValue(hive_hklm, "(Default)", value_str))

    subkey.add_subkey(subsubkey_name, subsubkey)
    clsid_key.add_subkey(subkey_name, subkey)
    hive_hklm.map_key(clsid_key_name, clsid_key)

    target_win_users.add_plugin(CLSIDPlugin)

    results = list(target_win_users.clsid())

    assert len(results) == 1

    result = results[0]

    assert result.regf_hive_path is None
    assert result.value == value_str



"""A pathlib.Path compatible implementation for dissect.target.

This allows for the majority of the pathlib.Path API to "just work" on dissect.target filesystems.

Most of this consists of subclassed internal classes with dissect.target specific patches,
but sometimes the change to a function is small, so the entire internal function is copied
and only a small part changed. To ease updating this code, the order of functions, comments
and code style is kept largely the same as the original pathlib.py.

Yes, we know, this is playing with fire and it can break on new CPython releases.

The implementation is split up in multiple files, one for each CPython version.
You're currently looking at the CPython 3.12 implementation.

Commit hash we're in sync with: f49221a

Notes:
    - CPython 3.12 changed a lot in preparation of proper subclassing, so our patches differ
      a lot from previous versions
    - Flavours don't really exist anymore, but since we kind of "multi-flavour" we need to emulate it
"""

from __future__ import annotations

import posixpath
import sys
from pathlib import Path, PurePath
from typing import IO, TYPE_CHECKING, ClassVar

from dissect.target import filesystem
from dissect.target.exceptions import FilesystemError, SymlinkRecursionError
from dissect.target.helpers import polypath
from dissect.target.helpers.compat import path_common

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem, FilesystemEntry
    from dissect.target.helpers.fsutil import stat_result


class _DissectFlavour:
    sep = "/"
    altsep = ""
    case_sensitive = False

    __variant_instances: ClassVar[dict[tuple[bool, str], _DissectFlavour]] = {}

    def __new__(cls, case_sensitive: bool = False, alt_separator: str = ""):
        idx = (case_sensitive, alt_separator)
        instance = cls.__variant_instances.get(idx, None)
        if instance is None:
            instance = object.__new__(cls)
            cls.__variant_instances[idx] = instance

        return instance

    def __init__(self, case_sensitive: bool = False, alt_separator: str = ""):
        self.altsep = alt_separator
        self.case_sensitive = case_sensitive

    def normcase(self, s: str) -> str:
        return s if self.case_sensitive else s.lower()

    splitdrive = staticmethod(posixpath.splitdrive)

    def splitroot(self, part: str) -> tuple[str, str, str]:
        return polypath.splitroot(part, alt_separator=self.altsep)

    def join(self, *args) -> str:
        return polypath.join(*args, alt_separator=self.altsep)

    # NOTE: Fallback implementation from older versions of pathlib.py
    def ismount(self, path: TargetPath) -> bool:
        # Need to exist and be a dir
        if not path.exists() or not path.is_dir():
            return False

        try:
            parent_dev = path.parent.stat().st_dev
        except FilesystemError:
            return False

        dev = path.stat().st_dev
        if dev != parent_dev:
            return True
        ino = path.stat().st_ino
        parent_ino = path.parent.stat().st_ino
        return ino == parent_ino

    isjunction = staticmethod(path_common.isjunction)

    samestat = staticmethod(posixpath.samestat)

    def isabs(self, path: str) -> bool:
        return polypath.isabs(path, alt_separator=self.altsep)

    realpath = staticmethod(path_common.realpath)


class PureDissectPath(PurePath):
    _fs: Filesystem
    _flavour = _DissectFlavour(case_sensitive=False)

    def __reduce__(self) -> tuple:
        raise TypeError("TargetPath pickling is currently not supported")

    def __init__(self, fs: Filesystem, *pathsegments):
        if not isinstance(fs, filesystem.Filesystem):
            raise TypeError(
                "invalid PureDissectPath initialization: missing filesystem, "
                "got {!r} (this might be a bug, please report)".format(fs, *pathsegments)
            )

        alt_separator = fs.alt_separator
        path_args = []
        for arg in pathsegments:
            if isinstance(arg, str):
                arg = polypath.normalize(arg, alt_separator=alt_separator)
            path_args.append(arg)

        super().__init__(*path_args)
        self._fs = fs
        self._flavour = _DissectFlavour(alt_separator=fs.alt_separator, case_sensitive=fs.case_sensitive)

    def with_segments(self, *pathsegments) -> Self:
        return type(self)(self._fs, *pathsegments)

    # NOTE: This is copied from pathlib.py but turned into an instance method so we get access to the correct flavour
    def _parse_path(self, path: str) -> tuple[str, str, list[str]]:
        if not path:
            return "", "", []
        sep = self._flavour.sep
        altsep = self._flavour.altsep
        if altsep:
            path = path.replace(altsep, sep)
        drv, root, rel = self._flavour.splitroot(path)
        if not root and drv.startswith(sep) and not drv.endswith(sep):
            drv_parts = drv.split(sep)
            if len(drv_parts) == 4 and drv_parts[2] not in "?.":
                # e.g. //server/share
                root = sep
            elif len(drv_parts) == 6:
                # e.g. //?/unc/server/share
                root = sep
        parsed = [sys.intern(str(x)) for x in rel.split(sep) if x and x != "."]
        return drv, root, parsed

    def is_reserved(self) -> bool:
        """Return True if the path contains one of the special names reserved
        by the system, if any."""
        return False


class TargetPath(Path, PureDissectPath):
    __slots__ = ("_direntry", "_entry")

    def get(self) -> FilesystemEntry:
        """Return the :class:`FilesystemEntry` for this path."""
        if not hasattr(self, "_entry"):
            self._entry = self._direntry.get() if hasattr(self, "_direntry") else self._fs.get(str(self))
        return self._entry

    def stat(self, *, follow_symlinks: bool = True) -> stat_result:
        """
        Return the result of the stat() system call on this path, like
        os.stat() does.
        """
        if follow_symlinks:
            return self.get().stat()
        return self.get().lstat()

    def open(
        self,
        mode: str = "rb",
        buffering: int = 0,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> IO:
        """Open file and return a stream.

        Supports a subset of features of the real pathlib.open/io.open.

        Note: in contrast to regular Python, the mode is binary by default. Text mode
        has to be explicitly specified. Buffering is also disabled by default.
        """
        return path_common.io_open(self, mode, buffering, encoding, errors, newline)

    def write_bytes(self, data: bytes) -> int:
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        raise NotImplementedError("TargetPath.write_bytes() is unsupported")

    def write_text(
        self, data: str, encoding: str | None = None, errors: str | None = None, newline: str | None = None
    ) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        raise NotImplementedError("TargetPath.write_text() is unsupported")

    def iterdir(self) -> Iterator[Self]:
        """Iterate over the files in this directory.  Does not yield any
        result for the special paths '.' and '..'.
        """
        for entry in path_common.scandir(self):
            if entry.name in {".", ".."}:
                # Yielding a path object for these makes little sense
                continue
            child_path = self._make_child_relpath(entry.name)
            child_path._direntry = entry
            yield child_path

    def _scandir(self) -> path_common._DissectScandirIterator:
        return path_common.scandir(self)

    @classmethod
    def cwd(cls) -> Self:
        """Return a new path pointing to the current working directory."""
        raise NotImplementedError("TargetPath.cwd() is unsupported")

    @classmethod
    def home(cls) -> Self:
        """Return a new path pointing to the user's home directory (as
        returned by os.path.expanduser('~')).
        """
        raise NotImplementedError("TargetPath.home() is unsupported")

    def absolute(self) -> Self:
        """Return an absolute version of this path by prepending the current
        working directory. No normalization or symlink resolution is performed.

        Use resolve() to get the canonical path to a file.
        """
        raise NotImplementedError("TargetPath.absolute() is unsupported in Dissect")

    # NOTE: We changed some of the error handling here to deal with our own exception types
    def resolve(self, strict: bool = False) -> Self:
        """
        Make the path absolute, resolving all symlinks on the way and also
        normalizing it.
        """

        s = self._flavour.realpath(self, strict=strict)
        p = self.with_segments(s)

        # In non-strict mode, realpath() doesn't raise on symlink loops.
        # Ensure we get an exception by calling stat()
        if not strict:
            try:
                p.stat()
            except FilesystemError as e:
                if isinstance(e, SymlinkRecursionError):
                    raise
        return p

    def owner(self) -> str:
        """
        Return the login name of the file owner.
        """
        raise NotImplementedError("TargetPath.owner() is unsupported")

    def group(self) -> str:
        """
        Return the group name of the file gid.
        """
        raise NotImplementedError("TargetPath.group() is unsupported")

    def readlink(self) -> Self:
        """
        Return the path to which the symbolic link points.
        """
        return self.with_segments(self.get().readlink())

    def touch(self, mode: int = 0o666, exist_ok: bool = True) -> None:
        """
        Create this file with the given access mode, if it doesn't exist.
        """
        raise NotImplementedError("TargetPath.touch() is unsupported")

    def mkdir(self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False) -> None:
        """
        Create a new directory at this given path.
        """
        raise NotImplementedError("TargetPath.mkdir() is unsupported")

    def chmod(self, mode: int, *, follow_symlinks: bool = True) -> None:
        """
        Change the permissions of the path, like os.chmod().
        """
        raise NotImplementedError("TargetPath.chmod() is unsupported")

    def lchmod(self, mode: int) -> None:
        """
        Like chmod(), except if the path points to a symlink, the symlink's
        permissions are changed, rather than its target's.
        """
        raise NotImplementedError("TargetPath.lchmod() is unsupported")

    def unlink(self, missing_ok: bool = False) -> None:
        """
        Remove this file or link.
        If the path is a directory, use rmdir() instead.
        """
        raise NotImplementedError("TargetPath.unlink() is unsupported")

    def rmdir(self) -> None:
        """
        Remove this directory.  The directory must be empty.
        """
        raise NotImplementedError("TargetPath.rmdir() is unsupported")

    def rename(self, target: str) -> Self:
        """
        Rename this path to the target path.

        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.

        Returns the new Path instance pointing to the target path.
        """
        raise NotImplementedError("TargetPath.rename() is unsupported")

    def replace(self, target: str) -> Self:
        """
        Rename this path to the target path, overwriting if that path exists.

        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.

        Returns the new Path instance pointing to the target path.
        """
        raise NotImplementedError("TargetPath.replace() is unsupported")

    def symlink_to(self, target: str, target_is_directory: bool = False) -> None:
        """
        Make this path a symlink pointing to the target path.
        Note the order of arguments (link, target) is the reverse of os.symlink.
        """
        raise NotImplementedError("TargetPath.symlink_to() is unsupported")

    def hardlink_to(self, target: str) -> None:
        """
        Make this path a hard link pointing to the same file as *target*.

        Note the order of arguments (self, target) is the reverse of os.link's.
        """
        raise NotImplementedError("TargetPath.hardlink_to() is unsupported")

    def expanduser(self) -> Self:
        """Return a new path with expanded ~ and ~user constructs
        (as returned by os.path.expanduser)
        """
        raise NotImplementedError("TargetPath.expanduser() is unsupported")



from __future__ import annotations

import io

from dissect.target import container
from dissect.target.containers.qcow2 import QCow2Container
from dissect.target.filesystem import VirtualFilesystem
from tests._utils import absolute_path


def test_qcow2_container() -> None:
    """Test that QCOW2 containers are properly opened.

    Generated with::

        echo "testdissecteqcow2" > small.txt
        qemu-img convert -f raw -O qcow2 small.txt small.qcow2
    """
    path = absolute_path("_data/containers/qcow2/small.qcow2")

    fh = container.open(path)
    assert isinstance(fh, QCow2Container)
    a = fh.read(20)
    # trailing \x00 are expected. This is the same when using qemu-nbd
    assert a == b"testdissecteqcow2\n\x00\x00"
    assert fh.tell() == 20
    fh.seek(0, whence=io.SEEK_SET)
    assert fh.read(20) == b"testdissecteqcow2\n\x00\x00"
    fh.close()


def test_qcow2_detect_path() -> None:
    """Test that QCOW2 containers are properly opened, when using the extension based matching."""
    vfs = VirtualFilesystem()
    vfs.map_file("small.qcow2", absolute_path("_data/containers/qcow2/small.qcow2"))
    fh = container.open(vfs.path("small.qcow2"))
    assert isinstance(fh, QCow2Container)
    a = fh.read(20)
    # trailing \x00 are expected. This is the same when using qemu-nbd
    assert a == b"testdissecteqcow2\n\x00\x00"
    assert fh.tell() == 20
    fh.seek(0, whence=io.SEEK_SET)
    assert fh.read(20) == b"testdissecteqcow2\n\x00\x00"



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, BinaryIO

from dissect.vmfs import lvm
from dissect.vmfs.c_lvm import c_lvm

from dissect.target.volume import LogicalVolumeSystem, Volume

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self


class VmfsVolumeSystem(LogicalVolumeSystem):
    __type__ = "vmfs"

    def __init__(self, fh: BinaryIO | list[BinaryIO], *args, **kwargs):
        self.lvm = lvm.LVM(fh)
        super().__init__(fh, *args, **kwargs)

    @classmethod
    def open_all(cls, volumes: list[BinaryIO]) -> Iterator[Self]:
        lvm_volumes = defaultdict(list)

        for vol in volumes:
            if not cls.detect_volume(vol):
                continue

            device = lvm.Device(vol)
            for lv_meta in device.volumes:
                lv_id = (bytes(lv_meta.volMeta.lvID.uuid), lv_meta.volMeta.lvID.snapID)
                lvm_volumes[lv_id].append(device)

        for devices in lvm_volumes.values():
            try:
                yield cls(devices)
            except Exception:  # noqa: PERF203
                continue

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        return VmfsVolumeSystem.detect_volume(fh)

    @staticmethod
    def _detect_volume(fh: BinaryIO) -> bool:
        fh.seek(c_lvm.LVM_DEV_HEADER_OFFSET)
        sector = fh.read(512)
        return int.from_bytes(sector[:4], "little") == c_lvm.LVM_MAGIC_NUMBER

    def _volumes(self) -> Iterator[Volume]:
        for i, volume in enumerate(self.lvm.volumes):
            yield Volume(
                fh=volume.open(),
                number=i + 1,
                offset=None,
                size=volume.size,
                vtype=None,
                name=volume.name,
                guid=volume.uuid,
                raw=volume,
                disk=self.disk,
                vs=self,
            )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.database.exception import Error as DBError
from dissect.database.sqlite3 import SQLite3
from dissect.util.ts import cocoatimestamp

from dissect.target.exceptions import FileNotFoundError, UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import export
from dissect.target.plugins.apps.browser.browser import (
    GENERIC_HISTORY_RECORD_FIELDS,
    BrowserPlugin,
    try_idna,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target


class SafariPlugin(BrowserPlugin):
    """Safari browser plugin."""

    __namespace__ = "safari"

    DIRS = (
        # macOS
        "Library/Safari",
    )

    BrowserHistoryRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/safari/history", GENERIC_HISTORY_RECORD_FIELDS
    )

    def __init__(self, target: Target):
        super().__init__(target)
        self.installs = list(self._find_installs())

    def _find_installs(self) -> Iterator[tuple[UserDetails, Path]]:
        for user_details in self.target.user_details.all_with_home():
            for directory in self.DIRS:
                install = user_details.home_path.joinpath(directory)
                if install.is_dir():
                    yield user_details, install

    def check_compatible(self) -> None:
        if not self.installs:
            raise UnsupportedPluginError("No Safari directories found on target")

    def _iter_db(self, filename: str) -> Iterator[tuple[UserDetails, Path, SQLite3]]:
        for user_details, install in self.installs:
            db_file = install.joinpath(filename)
            try:
                yield user_details, db_file, SQLite3(db_file)
            except FileNotFoundError:
                self.target.log.info("Could not find %s file: %s", filename, db_file)
            except DBError as e:
                self.target.log.warning("Could not open %s file: %s", filename, db_file)
                self.target.log.debug("", exc_info=e)

    @export(record=BrowserHistoryRecord)
    def history(self) -> Iterator[BrowserHistoryRecord]:
        """Return browser history records from Safari.

        Yields BrowserHistoryRecord with the following fields:

        .. code-block:: text

            ts (datetime): Visit timestamp.
            browser (string): The browser from which the records are generated from.
            id (varint): Record ID.
            url (uri): History URL.
            title (string): Page title.
            description (string): Page description.
            host (string): Hostname.
            visit_type (varint): Visit type.
            visit_count (varint): Amount of visits.
            hidden (boolean): Hidden value (synthesized visits).
            typed (boolean): Typed value.
            session (varint): Session value.
            from_visit (varint): Record ID of the redirect source visit.
            from_url (uri): URL of the redirect source visit.
            source (path): The source file of the history record.
        """
        for user_details, db_file, db in self._iter_db("History.db"):
            try:
                items = {row.id: row for row in db.table("history_items").rows()}
                visits = {}

                for row in db.table("history_visits").rows():
                    visits[row.id] = row
                    item = items[row.history_item]

                    from_visit = visits.get(row.redirect_source) if row.redirect_source else None
                    from_item = items.get(from_visit.history_item) if from_visit else None

                    yield self.BrowserHistoryRecord(
                        ts=cocoatimestamp(row.visit_time),
                        browser="safari",
                        id=row.id,
                        url=try_idna(item.url),
                        title=row.title,
                        description=None,
                        host=item.domain_expansion,
                        visit_type=None,
                        visit_count=item.visit_count,
                        hidden=bool(row.synthesized),
                        typed=None,
                        session=None,
                        from_visit=row.redirect_source or None,
                        from_url=try_idna(from_item.url) if from_item else None,
                        source=db_file,
                        _target=self.target,
                        _user=user_details.user if user_details else None,
                    )
            except DBError as e:  # noqa: PERF203
                self.target.log.warning("Error processing history file: %s", db_file)
                self.target.log.debug("", exc_info=e)



from __future__ import annotations

import pytest

from dissect.target.plugins.apps.shell.powershell import PowerShellHistoryPlugin
from tests._utils import absolute_path


@pytest.mark.parametrize(
    ("target", "fs", "target_file"),
    [
        (
            "target_win_users",
            "fs_win",
            "users\\John\\AppData\\Roaming\\Microsoft\\Windows\\PowerShell\\psreadline\\ConsoleHost_history.txt",
        ),
        ("target_unix_users", "fs_unix", "/root/.local/share/powershell/PSReadLine/ConsoleHost_history.txt"),
    ],
)
def test_powershell(target: str, fs: str, target_file: str, request: pytest.FixtureRequest) -> None:
    fs = request.getfixturevalue(fs)
    target = request.getfixturevalue(target)

    history_file = absolute_path("_data/plugins/os/windows/powershell/ConsoleHost_history.txt")
    fs.map_file(target_file, history_file)

    if target_file.startswith("users\\"):
        target_file = target_file.replace("users\\", "C:\\Users\\")

    target.add_plugin(PowerShellHistoryPlugin)

    records = list(target.powershell_history())

    assert len(records) == 4
    assert records[0].command == 'Write-Host "Hello World!"'
    assert records[0].order == 0
    assert str(records[0].source) == target_file



from __future__ import annotations

import io
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.helpers import keychain
from dissect.target.plugins.os.windows.dpapi.crypto import (
    _AES256,
    _DES3,
    _SHA1,
    _SHA512,
)
from dissect.target.plugins.os.windows.dpapi.dpapi import DPAPIBlob, DPAPIPlugin, MasterKeyFile
from tests._utils import absolute_path
from tests.conftest import add_win_user
from tests.plugins.os.windows.test__os import map_version_value
from tests.plugins.os.windows.test_lsa import (
    POLICY_KEY_PATH_NT5,
    POLICY_KEY_PATH_NT6,
    map_lsa_polkey,
    map_lsa_secrets,
    map_lsa_system_keys,
)

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.helpers.regutil import VirtualHive
    from dissect.target.target import Target

"""
You can use the following PowerShell code to generate user and SYSTEM DPAPI encrypted blobs.
This has been tested on every major Windows release since XP and works on PowerShell 1.0 and up.

.. code-block:: text

    "example text" | ConvertTo-SecureString -AsPlainText -Force | ConvertFrom-SecureString | Set-Content -Path out.txt

To create mock Windows targets you'll need the following additional data:
    - add_win_user: SID of the user the secret was generated by / for
    - map_lsa_system_keys: LSA system keys
    - map_lsa_polkey: LSA policy key
    - map_lsa_secrets: LSA DPAPI_SYSTEM encrypted secret
    - DPAPI master key that was used for encrypting the blob
    - The user's password or secret used to encrypt the DPAPI master key
"""


@pytest.mark.parametrize(
    ("user_type", "master_key_guid", "enc_data", "plaintext"),
    [
        pytest.param(
            "user",
            "f5638f3b-9632-480b-ba9c-c73bdabe9f86",
            (
                "01000000d08c9ddf0115d1118c7a00c04fc297eb010000003b8f63f532960b48"
                "ba9cc73bdabe9f8600000000020000000000106600000001000020000000d077"
                "d64b2d9bba2d3bc621a9a58d59de28aac032417f13ed1c8036503e4a3bce0000"
                "00000e8000000002000020000000043484aef65bad8467509d09906fe98601b4"
                "1cc04a7d417e83b0b00c02a81421500000004650eecc2f3bf3f6b95dafa29f1c"
                "2595f35a28b2f4a76c0a984e0aa65db83705517e5f98ce0de9f3a1abffe711f1"
                "b8129deab037dd790125542ba520f8f6ba0d863f36aea01b5e9222e98de1aeee"
                "9dbd400000005b938c2d3433effaaa23f04e5c272e8bd0877073a83cf43aed75"
                "82d22c59522856e2dfb6585bc82c4c66081d01f0b8fbfdc03641db3e6e195c47"
                "3940324b68b9"
            ),
            "this is a windows 10 user secret!",
            id="user",
        ),
        pytest.param(
            "system",
            "59873e79-ca24-459a-8b17-a5dcd82c6ea1",
            (
                "01000000d08c9ddf0115d1118c7a00c04fc297eb01000000793e875924ca9a45"
                "8b17a5dcd82c6ea100000000020000000000106600000001000020000000bf65"
                "33a2eaa9c0393173bc5b8b53c33bb6273b8b86f6586a64a3cb5a2dbab7190000"
                "00000e80000000020000200000006c29f425bf48b2bf66d9fe2c6c406a48c41a"
                "77c90e911fe6576bc9811553ef1d50000000e5bf02ccca4cc273aa5dfacfa64b"
                "c3ab746f59de5699943edbe0fdb6b389c73e0541589d4ac9f37cf4488c44bf0f"
                "7fc656dc2bcad919c112a8cda9b54ede0d31ea260067f5540c0eecb3322682e2"
                "0b6b400000009a07d60ab8ef3632b8f54009ece9599beba174626c8fe0dd8642"
                "bd207dde9a6c34f1f611eccb6606d1e1ea10a406930b4e11087031dbe083a66e"
                "e5db2b95ff71"
            ),
            "this is a windows 10 system secret!",
            id="system",
        ),
    ],
)
def test_dpapi_decrypt_blob_win_10(
    target_win: Target,
    fs_win: VirtualFilesystem,
    hive_hklm: VirtualHive,
    hive_hku: VirtualHive,
    user_type: str,
    master_key_guid: str,
    enc_data: str,
    plaintext: str,
) -> None:
    """Test decrypting a windows 10 dpapi system and user blob."""

    add_win_user(
        hive_hklm,
        hive_hku,
        target_win,
        sid="S-1-5-21-1555088973-3915578919-3195617063-1003",
        home="c:\\users\\user",
    )

    map_lsa_system_keys(hive_hklm, {"Data": "3b03ce4f", "GBG": "7bf5e093", "JD": "51fe8c74", "Skew1": "3e14f655"})
    map_lsa_polkey(
        hive_hklm,
        POLICY_KEY_PATH_NT6,
        bytes.fromhex(
            "00000001ecffe17b2a997440aa939adbff26f1fc030000000000000061f61040"
            "72b7b9117289a891bc1edb9fa83fafae5d8648b8674825dc9bdcf1bf5ffbe591"
            "27fd3a1baca6c5447530574c27a41ab721b13e54a9af857e682b8458859c0fab"
            "55461a8f1b2b8c12d8f186a80991012e2d3ede5b1167554c726e404ed0eb96f6"
            "488b90ece5b0b3ef9bf244f35dd65a30b2b450c4e227838810b631564b89300e"
            "d4355040f23890f8b6b8eab8"
        ),
    )
    map_lsa_secrets(
        hive_hklm,
        {
            "DPAPI_SYSTEM": bytes.fromhex(
                "00000001001f9b85984f68a8ed3e9d44dbd5b79c0300000000000000835b4ff6"
                "6e74154ffab75afd31a2860616c87411bc97d368a068fca62d1564ae1396f88e"
                "06de87b5e5632c668538ee36c75f67cee98d49ebc1e88fa7e9be16144af31e8e"
                "5fd78329279e50d792e8a6b35a59cb55016748ecd8e12f148b1d32b3"
            )
        },
    )

    # user master key
    fs_win.map_file(
        "Users/user/AppData/Roaming/Microsoft/Protect/S-1-5-21-1555088973-3915578919-3195617063-1003/f5638f3b-9632-480b-ba9c-c73bdabe9f86",
        absolute_path("_data/plugins/os/windows/dpapi/master_keys/f5638f3b-9632-480b-ba9c-c73bdabe9f86"),
    )

    # system master key
    fs_win.map_file(
        "Windows/System32/Microsoft/Protect/S-1-5-18/User/59873e79-ca24-459a-8b17-a5dcd82c6ea1",
        absolute_path("_data/plugins/os/windows/dpapi/master_keys/59873e79-ca24-459a-8b17-a5dcd82c6ea1"),
    )

    map_version_value(target_win, "CurrentVersion", 10.0)
    target_win.add_plugin(DPAPIPlugin)

    # test parsing dpapi system blob
    enc_data = bytes.fromhex(enc_data)
    blob = DPAPIBlob(enc_data)
    assert blob.version == 1
    assert blob.provider == "df9d8cd0-1501-11d1-8c7a-00c04fc297eb"
    assert blob.guid == master_key_guid
    assert isinstance(blob.cipher_algorithm, _AES256)
    assert isinstance(blob.hash_algorithm, _SHA512)

    if user_type == "user":
        # test decrypting dpapi user blob
        keychain.register_key(
            key_type=keychain.KeyType.PASSPHRASE,
            value="password",
            identifier=None,
            provider="user",
        )
        decrypted = target_win.dpapi.decrypt_user_blob(enc_data, sid="S-1-5-21-1555088973-3915578919-3195617063-1003")
    elif user_type == "system":
        # test decrypting dpapi system blob
        decrypted = target_win.dpapi.decrypt_system_blob(enc_data)

    assert decrypted.decode("utf-16-le") == plaintext


@pytest.mark.parametrize(
    ("user_type", "master_key_guid", "enc_data", "plaintext"),
    [
        pytest.param(
            "user",
            "835419eb-ce4a-4ae3-a72a-35c1383a6519",
            (
                "01000000d08c9ddf0115d1118c7a00c04fc297eb01000000eb1954834acee34a"
                "a72a35c1383a65190000000002000000000010660000000100002000000093f6"
                "c9a6ac508949a451976699e226145277e67f88fd1bea46576df1f8386fc10000"
                "00000e80000000020000200000003ff5eae86a8f83bb0517a0706535b1107d9f"
                "e72682ca4f68b7dfbc56939f0ef530000000ab26fca786c57987910bd313da0c"
                "0426b42d139d4b9904198f45ca51996eb6c888b9ba50d5787d18fb56826db726"
                "e8fd400000009ee887a2f48d5812921dc870691b49b94d368a03c4f55cdceb00"
                "fa5c427b7c4f151d5a6de35b8902d18698905f0abd89e03bbc948906977ff5c4"
                "e40d07649869"
            ),
            "this is a user secret!",
            id="user",
        ),
        pytest.param(
            "system",
            "ab15130c-61cd-4065-be48-faba655f80d1",
            (
                "01000000d08c9ddf0115d1118c7a00c04fc297eb010000000c1315abcd616540"
                "be48faba655f80d100000000020000000000106600000001000020000000511a"
                "83f59218606393ea986893f8d39156eed29ad43c110f171ad5e1aa63a8a40000"
                "00000e800000000200002000000029cb546db6b46d4407ead0366bd7a5c52e97"
                "eebc1291128d7d19bfdf9736449a400000005ec4e78e95d88d4ca584a23e506e"
                "e0677aec069c38631dd9c7d49559cfeb60ed4962fab7bfd686aa06718c9be667"
                "d1126169baa4c5d985dcb029c2d3febd6e3f400000003829a0d94657938a8f29"
                "a399931f9bddc2208f8dc9e7940b494147502793dcb38c4265943d77d8aac204"
                "a80b813315de746d3334e7abb3bddae312a87b132384"
            ),
            "this is a system secret!",
            id="system",
        ),
    ],
)
def test_dpapi_decrypt_blob_win_7(
    target_win: Target,
    fs_win: VirtualFilesystem,
    hive_hklm: VirtualHive,
    hive_hku: VirtualHive,
    user_type: str,
    master_key_guid: str,
    enc_data: str,
    plaintext: str,
) -> None:
    """Test decrypting a windows 7 user and system blob."""

    add_win_user(hive_hklm, hive_hku, target_win, sid="S-1-5-18", home="%systemroot%\\system32\\config\\systemprofile")
    add_win_user(
        hive_hklm,
        hive_hku,
        target_win,
        sid="S-1-5-21-2423314426-1454842194-222297899-1000",
        home="C:\\Users\\user",
    )

    map_lsa_system_keys(hive_hklm, {"Data": "7ea9249d", "GBG": "32007b89", "JD": "1f884e65", "Skew1": "f5b653a9"})
    map_lsa_polkey(
        hive_hklm,
        POLICY_KEY_PATH_NT6,
        bytes.fromhex(
            "00000001ecffe17b2a997440aa939adbff26f1fc0300000000000000866bf805"
            "f051a5a0a4aa4fc6e8d460b9c2106b47adf9096f610d84da45a9c092cba0787d"
            "1bc4d161ebcec3f16ac86ba57de34aec8f5c1fc950998d3ad05160b2aa9df553"
            "5ec99c36cd00d6cbef7becc50d4e320a6722fbb6f4fd26cdaf9a78205a7b8481"
            "c6e51f6431081d602fc1e77d24f38d0c0e3eb3ceca952efbe2635960e9ac1995"
            "e1a0380659bf1a8c5226b7d1"
        ),
    )

    map_lsa_secrets(
        hive_hklm,
        {
            "DPAPI_SYSTEM": bytes.fromhex(
                "00000001f9da1883a8c8a2e5f8c9e783b65ef8e90300000000000000f293e237"
                "48961ef80e666f05317889b6d86143400b3282cc087dcc8a6a3f6213d6e945c8"
                "e323903818451e15d09201a50ed43c629acdb20d23424bc2625acfdf9b94ed9f"
                "3868463f6c7adc8736a8cc4e70e115c621c2d482f07d501d21af32b9"
            ),
            "DefaultPassword": bytes.fromhex(
                "00000001f9da1883a8c8a2e5f8c9e783b65ef8e903000000000000003f0e05667"
                "69d976dcabfc1c10571bc0c6a6ca5032a258c28c74fc7646680bceafcb70fd7de"
                "e974854835260283fcb0dc3af665763b7757049c01730212a5cace00000000000"
                "000000000000000000000"
            ),
        },
    )

    # user master key
    fs_win.map_file(
        "Users/user/AppData/Roaming/Microsoft/Protect/S-1-5-21-2423314426-1454842194-222297899-1000/835419eb-ce4a-4ae3-a72a-35c1383a6519",
        absolute_path("_data/plugins/os/windows/dpapi/master_keys/835419eb-ce4a-4ae3-a72a-35c1383a6519"),
    )

    # system master key
    fs_win.map_file(
        "Windows/System32/Microsoft/Protect/S-1-5-18/User/ab15130c-61cd-4065-be48-faba655f80d1",
        absolute_path("_data/plugins/os/windows/dpapi/master_keys/ab15130c-61cd-4065-be48-faba655f80d1"),
    )

    map_version_value(target_win, "CurrentVersion", 6.1)
    target_win.add_plugin(DPAPIPlugin)

    # test dpapi blob parsing
    enc_data = bytes.fromhex(enc_data)
    blob = DPAPIBlob(enc_data)
    assert blob.version == 1
    assert blob.guid == master_key_guid
    assert blob.provider == "df9d8cd0-1501-11d1-8c7a-00c04fc297eb"
    assert isinstance(blob.cipher_algorithm, _AES256)
    assert isinstance(blob.hash_algorithm, _SHA512)

    if user_type == "user":
        # test decrypting dpapi blob using user DefaultPassword lsa secret
        decrypted = target_win.dpapi.decrypt_user_blob(enc_data, sid="S-1-5-21-2423314426-1454842194-222297899-1000")

    elif user_type == "system":
        # test decrypting dpapi blob using DPAPI_SYSTEM lsa secret
        decrypted = target_win.dpapi.decrypt_system_blob(enc_data)

    assert decrypted.decode("utf-16-le") == plaintext


@pytest.mark.parametrize(
    ("user_type", "master_key_guid", "enc_data", "plaintext"),
    [
        pytest.param(
            "user",
            "8bd87dd9-10fa-40b5-8614-5d0a7e8911c5",
            (
                "01000000d08c9ddf0115d1118c7a00c04fc297eb01000000d97dd88bfa10b540"
                "86145d0a7e8911c50000000002000000000003660000a800000010000000c58b"
                "6464cb47b91908f333a4947bc9770000000004800000a000000010000000fa89"
                "775be164fdbe5de900710788dd075000000020a462e392783abaadaec0b42b74"
                "dd729f19e7bb4cb6bed585e4f11d332d52a5e35d998d63b350172eae76a595e8"
                "cc084dff8389900fd0a560d192d2e982b83382bc39bc37527c7eee989624fef5"
                "91941400000070e42b6873f1372abd5dbaced1e97dd0fbeee5c0"
            ),
            "this is a windows vista user secret!",
            id="user",
        ),
        pytest.param(
            "system",
            "9ed75ec3-d074-4a7a-a428-1b70188cce4c",
            (
                "01000000d08c9ddf0115d1118c7a00c04fc297eb01000000c35ed79e74d07a4a"
                "a4281b70188cce4c0000000002000000000003660000a80000001000000003e3"
                "f985868f3a9e6e1628672bc702610000000004800000a000000010000000ffa0"
                "2bf52d9cbfd640921b60560e586c50000000b0705dd3c0975c5a21de7d77785b"
                "756b0a6ee22fe7ee080730fa213630186ccc96c98d6cfd4903cd0ba0008016f9"
                "99cc6d3ca8375e42561b1289e3db3351784dd9b268163998ba61b190abf50bf1"
                "bea5140000003ec445826ec3c1dce592a6b15d467290ca1f4068"
            ),
            "this is a windows vista system secret!",
            id="system",
        ),
    ],
)
def test_dpapi_decrypt_blob_win_vista(
    target_win: Target,
    fs_win: VirtualFilesystem,
    hive_hklm: VirtualHive,
    hive_hku: VirtualHive,
    user_type: str,
    master_key_guid: str,
    enc_data: str,
    plaintext: str,
) -> None:
    """Test decrypting a windows vista user and system blob."""
    add_win_user(hive_hklm, hive_hku, target_win, sid="S-1-5-18", home="%systemroot%\\system32\\config\\systemprofile")
    add_win_user(
        hive_hklm, hive_hku, target_win, sid="S-1-5-21-3694181269-2347385980-100747364-1000", home="C:\\Users\\user"
    )

    map_lsa_system_keys(
        hive_hklm,
        {
            "Data": "c948a244",
            "GBG": "f72a0d84",
            "JD": "4974e288",
            "Skew1": "485dc08b",
        },
    )
    map_lsa_polkey(
        hive_hklm,
        POLICY_KEY_PATH_NT6,
        bytes.fromhex(
            "00000001ecffe17b2a997440aa939adbff26f1fc03000000000000007cf4b483"
            "1c3fb8be797fc9f30e4226bf3526520458eacb05b3dd0db072b07d25683ee35e"
            "69a7461ebd65593ffd3dc0e5be31b31deec2573dc6367258783e0394d1b55a2b"
            "e6d7931881eeb391da41a062e17419a35e1b0000b2717a362cbc81e89b54f85c"
            "801a1a46a492491032e7055aa46730d3d78530834547b1c6ca42da34d5dacac9"
            "b9ce6013c3486f925401a97b"
        ),
    )
    map_lsa_secrets(
        hive_hklm,
        {
            "DPAPI_SYSTEM": bytes.fromhex(
                "00000001340de2496d244a12a38743bdb773cfd203000000000000006d926cee"
                "7ae8a462255cfbd47b7eae097dd9104d61591cca7e7973eecbb5bbc468f375856"
                "a63f6d5caa1f8f796220d12117cf455f85219a9f8ab0c04fd553529e257baa17b"
                "b3e81c5cc62af835e7332565b38b1a97f590df056044ad38c2e88f"
            ),
            "DefaultPassword": bytes.fromhex(
                "00000001340de2496d244a12a38743bdb773cfd2030000000000000094eb76f9"
                "2b49a37b98cd19f98ab5372eab0e18dbdc6e36729f31961a77197c69c4115cd3"
                "7dd0cf2e7a002453c9280bb700811db5fc09517bc7eb387f8aebee6a00000000"
                "000000000000000000000000"
            ),
        },
    )

    # user master key
    fs_win.map_file(
        "Users/user/AppData/Roaming/Microsoft/Protect/S-1-5-21-3694181269-2347385980-100747364-1000/8bd87dd9-10fa-40b5-8614-5d0a7e8911c5",
        absolute_path("_data/plugins/os/windows/dpapi/master_keys/8bd87dd9-10fa-40b5-8614-5d0a7e8911c5"),
    )

    # system master key
    fs_win.map_file(
        "Windows/System32/Microsoft/Protect/S-1-5-18/User/9ed75ec3-d074-4a7a-a428-1b70188cce4c",
        absolute_path("_data/plugins/os/windows/dpapi/master_keys/9ed75ec3-d074-4a7a-a428-1b70188cce4c"),
    )

    map_version_value(target_win, "CurrentVersion", 6.0)
    target_win.add_plugin(DPAPIPlugin)

    enc_data = bytes.fromhex(enc_data)

    blob = DPAPIBlob(enc_data)
    assert blob.version == 1
    assert blob.provider == "df9d8cd0-1501-11d1-8c7a-00c04fc297eb"
    assert blob.guid == master_key_guid
    assert blob.description == "\x00"
    assert isinstance(blob.cipher_algorithm, _DES3)
    assert isinstance(blob.hash_algorithm, _SHA1)

    if user_type == "user":
        # test decrypting dpapi blob using user DefaultPassword lsa secret
        decrypted = target_win.dpapi.decrypt_user_blob(enc_data, username="user")
    elif user_type == "system":
        # test decrypting dpapi blob using DPAPI_SYSTEM lsa secret
        decrypted = target_win.dpapi.decrypt_system_blob(enc_data)

    assert decrypted.decode("utf-16-le") == plaintext


@pytest.mark.parametrize(
    ("user_type", "master_key_guid", "enc_data", "plaintext"),
    [
        pytest.param(
            "user",
            "e1245b49-43d9-465f-b161-eab88c27620d",
            (
                "01000000d08c9ddf0115d1118c7a00c04fc297eb01000000495b24e1d9435f46"
                "b161eab88c27620d0000000002000000000003660000a800000010000000631f"
                "9403ae62478940353bc54274d0bd0000000004800000a00000001000000081aa"
                "f6b9a2f370073f8f869269607dd348000000312f2f35f6751e099685f93e43b5"
                "9f77e7bd0202e0efb2ed1412a5f9605a7b1229e379fcb6b6de141eee7e7eaa14"
                "ff14899edb6e805903ddd81c2b1b65cfa8246ce3da28dd3a1c4214000000d9d1"
                "f626a57c4b1c4dd15267560b945c4c7e9124"
            ),
            "this is a windows xp user secret!",
            id="user",
        ),
        pytest.param(
            "system",
            "2ee95631-ed37-43f9-8a08-378924a60de7",
            (
                "01000000d08c9ddf0115d1118c7a00c04fc297eb010000003156e92e37edf943"
                "8a08378924a60de70000000002000000000003660000a8000000100000000e23"
                "be00ab6ef398df788e3fbcafbdcc0000000004800000a0000000100000005af1"
                "9111a4f165ffebb7bbf4133f634b4800000090605c23c6a9225592f6e6e75aa2"
                "32566a2d36037ebdacef182aaa677c0b64e541ca1c59732125b0149f0031c48f"
                "3d69c2fde832fa304d0b7af22b6fda2a6228507988d7cfe49dbf140000003c75"
                "50be593b511450387842969ee1217fc07870"
            ),
            "this is a windows xp system secret!",
            id="system",
        ),
    ],
)
def test_dpapi_decrypt_blob_win_xp(
    target_win: Target,
    fs_win: VirtualFilesystem,
    hive_hklm: VirtualHive,
    hive_hku: VirtualHive,
    user_type: str,
    master_key_guid: str,
    enc_data: str,
    plaintext: str,
) -> None:
    """Test decrypting a windows xp user and system blob."""

    add_win_user(hive_hklm, hive_hku, target_win, sid="S-1-5-18", home="%systemroot%\\system32\\config\\systemprofile")
    add_win_user(
        hive_hklm,
        hive_hku,
        target_win,
        sid="S-1-5-21-1960408961-57989841-725345543-1003",
        home="C:\\Documents and Settings\\user",
    )

    map_lsa_system_keys(hive_hklm, {"Data": "0e1eaddb", "GBG": "36ec555b", "JD": "8397b619", "Skew1": "9dfad6c8"})
    map_lsa_polkey(
        hive_hklm,
        POLICY_KEY_PATH_NT5,
        bytes.fromhex(
            "0100000001000000000000003e4253c6ab72f73306ffe45a33a4ec5d47d21483"
            "85be6036b4c5bab2666d7f5793f23fed8513577b79abddb40eaa92c7a19b34ab"
            "b9c134e88bf5dd2b2fbf6356"
        ),
    )

    map_lsa_secrets(
        hive_hklm,
        {
            "DPAPI_SYSTEM": bytes.fromhex(
                "3800000038000020acb90c00a26440bdf0382bdbcb5d51a743cb4cca01cfecb3"
                "202848e00dfae25f8f0f3f4710ff37828532b5f600e97f6a00528fd8e0137f49"
                "6fd2d905"
            ),
        },
    )

    # user master key
    fs_win.map_file(
        "Documents and Settings/user/Application Data/Microsoft/Protect/S-1-5-21-1960408961-57989841-725345543-1003/e1245b49-43d9-465f-b161-eab88c27620d",  # noqa: E501
        absolute_path("_data/plugins/os/windows/dpapi/master_keys/e1245b49-43d9-465f-b161-eab88c27620d"),
    )

    # system master key
    fs_win.map_file(
        "Windows/System32/Microsoft/Protect/S-1-5-18/User/2ee95631-ed37-43f9-8a08-378924a60de7",
        absolute_path("_data/plugins/os/windows/dpapi/master_keys/2ee95631-ed37-43f9-8a08-378924a60de7"),
    )

    map_version_value(target_win, "CurrentVersion", 5.1)
    target_win.add_plugin(DPAPIPlugin)

    # make sure the user master key is decrypted without providing a password (empty keyprovider)
    user_mk = target_win.dpapi.master_keys.get("S-1-5-21-1960408961-57989841-725345543-1003", {}).get(
        "e1245b49-43d9-465f-b161-eab88c27620d"
    )
    assert isinstance(user_mk, MasterKeyFile)
    assert user_mk.decrypted

    enc_data = bytes.fromhex(enc_data)

    blob = DPAPIBlob(enc_data)
    assert blob.version == 1
    assert blob.provider == "df9d8cd0-1501-11d1-8c7a-00c04fc297eb"
    assert blob.guid == master_key_guid
    assert blob.description == "\x00"
    assert isinstance(blob.cipher_algorithm, _DES3)
    assert isinstance(blob.hash_algorithm, _SHA1)

    if user_type == "user":
        # test decrypting dpapi blob using user master key
        decrypted = target_win.dpapi.decrypt_user_blob(enc_data, username="user")

    elif user_type == "system":
        # test decrypting dpapi blob using DPAPI_SYSTEM lsa secret
        decrypted = target_win.dpapi.decrypt_system_blob(enc_data)

    assert decrypted.decode("utf-16-le") == plaintext


def test_dpapi_master_keys_deduplicate(
    target_win: Target, fs_win: VirtualFilesystem, hive_hklm: VirtualHive, hive_hku: VirtualHive
) -> None:
    """Test if we correctly deduplicate master keys."""

    add_win_user(
        hive_hklm,
        hive_hku,
        target_win,
        sid="S-1-5-21-1555088973-3915578919-3195617063-1003",
        home="c:\\users\\user",
    )

    # Mock "Application Data" -> "AppData/Roaming" symlink
    fs_win.symlink("/sysvol/Users/user/AppData/Roaming", "/Users/user/Application Data")

    fs_win.map_file(
        "Users/user/AppData/Roaming/Microsoft/Protect/S-1-5-21-1555088973-3915578919-3195617063-1003/f5638f3b-9632-480b-ba9c-c73bdabe9f86",
        absolute_path("_data/plugins/os/windows/dpapi/master_keys/f5638f3b-9632-480b-ba9c-c73bdabe9f86"),
    )

    plugin = target_win.add_plugin(DPAPIPlugin, check_compatible=False)

    with patch("dissect.target.plugins.os.windows.dpapi.dpapi.MasterKeyFile") as MasterKeyFileMock:
        dict(plugin.master_keys)
        MasterKeyFileMock.assert_called_once()


def test_dpapi_master_key_decrypt_with_hash(
    target_win: Target, fs_win: VirtualFilesystem, hive_hklm: VirtualHive, hive_hku: VirtualHive
) -> None:
    """Test that 'master_keys' are decrypted first with a password and then with a hash."""
    # Setup: create a fake master key file and hash
    sid = "S-1-5-21-1111111111-2222222222-3333333333-1001"
    mk_guid = "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee"
    mk_hash = "0123456789abcdef0123456789abcdef"  # 16 bytes hex string

    add_win_user(hive_hklm, hive_hku, target_win, sid=sid, home="C:\\Users\\user")
    map_lsa_system_keys(
        hive_hklm,
        {"Data": "deadbeef", "GBG": "cafebabe", "JD": "feedface", "Skew1": "badc0de1"},
    )

    # Register the hash in the keychain
    class MyStr(str):
        __slots__ = ()

    keychain.register_key(
        key_type=keychain.KeyType.PASSPHRASE,
        value=MyStr(mk_hash),
        identifier=None,
        provider="user",
    )

    # Map a dummy master key file using an in-memory BytesIO
    fs_win.map_file_fh(f"Users/user/AppData/Roaming/Microsoft/Protect/{sid}/{mk_guid}", io.BytesIO(b"\x00" * 128))

    map_version_value(target_win, "CurrentVersion", 10.0)
    target_win.add_plugin(DPAPIPlugin)

    # Patch MasterKeyFile to track decrypted state in a closure
    decrypted_state = False

    def fake_decrypt_with_password(self: MasterKeyFile, *args, **kwargs) -> bool:
        nonlocal decrypted_state
        decrypted_state = False
        return False

    def fake_decrypt_with_hash(self: MasterKeyFile, user_sid: str, password_hash: bytes) -> bool:
        nonlocal decrypted_state
        decrypted_state = len(password_hash) > 0
        return decrypted_state

    def decrypted_getter(self: MasterKeyFile) -> bool:
        return decrypted_state

    with (
        patch.object(MasterKeyFile, "decrypt_with_password", new=fake_decrypt_with_password),
        patch.object(MasterKeyFile, "decrypt_with_hash", new=fake_decrypt_with_hash),
        patch.object(MasterKeyFile, "decrypted", new=property(decrypted_getter)),
    ):
        # 'master_keys' is the subject under test: accessing it should trigger the PyPy workaround
        keys = target_win.dpapi.master_keys
        assert sid in keys
        assert mk_guid in keys[sid]
        assert decrypted_state



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.cyber import CyberLoader
from dissect.target.loaders.tar import TarLoader
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``CyberLoader`` when opening a ``Target``."""
    file_path = absolute_path("_data/loaders/tar/test-archive.tar")
    path = f"cyber://{file_path}"

    with patch("dissect.target.loaders.cyber.cyber"):
        target = opener(path)
        assert isinstance(target._loader, CyberLoader)
        assert isinstance(target._loader._real, TarLoader)
        assert target.path == file_path


def test_loader() -> None:
    """Test that ``CyberLoader`` correctly loads a Cyber file."""
    path = f"cyber://{absolute_path('_data/loaders/tar/test-archive.tar')}"

    loader = loader_open(path)
    assert isinstance(loader, CyberLoader)
    assert isinstance(loader._real, TarLoader)

    t = Target()
    with patch("dissect.target.loaders.cyber.cyber"):
        loader.map(t)

    assert "cyber" in t.props



from __future__ import annotations

import platform
import re
from functools import cache, cached_property
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import InternalPlugin
from dissect.target.plugins.os.windows.dpapi.blob import Blob as DPAPIBlob
from dissect.target.plugins.os.windows.dpapi.master_key import CredSystem, MasterKeyFile

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target


class DPAPIPlugin(InternalPlugin):
    """Windows Data Protection API (DPAPI) plugin.

    References:
        - Reversing ``Crypt32.dll``
        - https://learn.microsoft.com/en-us/windows/win32/api/dpapi/
        - https://github.com/fortra/impacket/blob/master/examples/dpapi.py
        - https://github.com/tijldeneut/DPAPIck3
        - https://www.passcape.com/index.php?section=docsys&cmd=details&id=28
    """

    __namespace__ = "dpapi"

    RE_MASTER_KEY = re.compile("^[0-9a-f]{8}(?:-[0-9a-f]{4}){3}-[0-9a-f]{12}$")
    SYSTEM_SID = "S-1-5-18"

    def __init__(self, target: Target):
        super().__init__(target)
        self._seen_mks = set()
        self.keychain = cache(self.keychain)

    def check_compatible(self) -> None:
        if not self.target.has_function("lsa"):
            raise UnsupportedPluginError("Windows registry and LSA plugins are required for DPAPI decryption")

    def keychain(self) -> set:
        return set(self.target.dpapi.keyprovider.keys())

    @cached_property
    def master_keys(self) -> dict[str, dict[str, MasterKeyFile]]:
        """Returns dict of found DPAPI master keys on the Windows target for SYSTEM and regular users."""
        master_keys = {}

        # Search for SYSTEM master keys
        master_keys[self.SYSTEM_SID] = {}

        system_master_key_path = self.target.resolve(f"%windir%/System32/Microsoft/Protect/{self.SYSTEM_SID}")
        system_user_master_key_path = system_master_key_path.joinpath("User")

        for dir in [system_master_key_path, system_user_master_key_path]:
            user_mks = self._load_master_keys_from_path(self.SYSTEM_SID, dir)
            master_keys[self.SYSTEM_SID].update(user_mks)

        # Search for user master keys, generally located at $HOME/AppData/Roaming/Microsoft/Protect/{user_sid}/{mk_guid}
        PROTECT_DIRS = [
            # Windows Vista and newer
            "AppData/Roaming/Microsoft/Protect",
            # Windows XP
            "Application Data/Microsoft/Protect",
        ]

        for user in self.target.user_details.all_with_home():
            sid = user.user.sid
            master_keys.setdefault(sid, {})

            for protect_dir in PROTECT_DIRS:
                path = user.home_path.joinpath(protect_dir).joinpath(sid)
                if user_mks := self._load_master_keys_from_path(sid, path):
                    master_keys[sid] |= user_mks

        return master_keys

    def _load_master_keys_from_path(self, sid: str, path: Path) -> Iterator[tuple[str, MasterKeyFile]]:
        """Iterate over the provided ``path`` and search for master key files for the given user SID."""

        if not path.exists():
            self.target.log.info("Unable to load master keys from path as it does not exist: %s", path)
            return

        for file in path.iterdir():
            if not self.RE_MASTER_KEY.findall(file.name):
                continue

            if (file.name, sid) in self._seen_mks:
                continue

            self._seen_mks.add((file.name, sid))

            with file.open() as fh:
                mkf = MasterKeyFile(fh)

            # Decrypt SYSTEM master key using the DPAPI_SYSTEM LSA secret.
            if sid == self.SYSTEM_SID:
                if "DPAPI_SYSTEM" not in self.target.lsa._secrets:
                    self.target.log.warning("Unable to decrypt SYSTEM master key: LSA secret missing")
                    continue

                # Windows XP or Windows Vista and newer
                secret_offset = 8 if float(self.target.ntversion) < 6.0 else 16

                dpapi_system = CredSystem(self.target.lsa._secrets["DPAPI_SYSTEM"][secret_offset:])
                mkf.decrypt_with_key(dpapi_system.machine_key)
                mkf.decrypt_with_key(dpapi_system.user_key)

                # Decrypting the System master key should always succeed
                if not mkf.decrypted:
                    self.target.log.error("Failed to decrypt SYSTEM master key!")
                    continue

                yield file.name, mkf

            # Decrypt user master key
            else:
                # Iterate over every master key password we have from the keychain
                for provider, mk_pass in self.keychain():
                    try:
                        if mkf.decrypt_with_password(sid, mk_pass):
                            self.target.log.info(
                                "Decrypted SID %s master key %s with password '%s' from provider %s",
                                sid,
                                file,
                                mk_pass,
                                provider,
                            )
                            break
                    except ValueError:
                        pass

                    try:
                        # NOTE: This is a workaround for a PyPy bug
                        # `bytes.fromhex` errors on subclasses of `str`
                        # (see https://github.com/pypy/pypy/issues/5327)
                        if platform.python_implementation() == "PyPy":
                            mk_pass = str(mk_pass)

                        if mkf.decrypt_with_hash(sid, bytes.fromhex(mk_pass)):
                            self.target.log.info(
                                "Decrypted SID %s master key %s with hash '%s' from provider %s",
                                sid,
                                file,
                                mk_pass,
                                provider,
                            )
                            break
                    except ValueError:
                        pass

                if not mkf.decrypted:
                    self.target.log.warning("Could not decrypt master key '%s' for SID '%s'", file.name, sid)

                yield file.name, mkf

    @cached_property
    def _users(self) -> dict[str, str]:
        """Cached map of username to SID."""
        return {user.name: user.sid for user in self.target.users()}

    def decrypt_system_blob(self, data: bytes, **kwargs) -> bytes:
        """Decrypt the given bytes using the SYSTEM master key.

        Args:
            data: Bytes of DPAPI system blob to decrypt.
            **kwargs: Arbitrary named arguments to pass to :meth:`DPAPIBlob.decrypt <dissect.target.plugins.os.windows.dpapi.blob.Blob.decrypt>` function.

        Raises:
            ValueError: When conditions to decrypt are not met or if decrypting failed.

        Returns:
            Decrypted bytes.
        """  # noqa: E501
        return self.decrypt_user_blob(data, sid=self.SYSTEM_SID, **kwargs)

    def decrypt_user_blob(self, data: bytes, username: str | None = None, sid: str | None = None, **kwargs) -> bytes:
        """Decrypt the given bytes using the master key of the given SID or username.

        Args:
            data: Bytes of DPAPI blob to decrypt.
            username: Username of the owner of the DPAPI blob.
            sid: SID of the owner of the DPAPI blob.
            **kwargs: Arbitrary named arguments to pass to :meth:`DPAPIBlob.decrypt <dissect.target.plugins.os.windows.dpapi.blob.Blob.decrypt>` function.

        Raises:
            ValueError: When conditions to decrypt are not met or if decrypting failed.

        Returns:
            Decrypted bytes.
        """  # noqa: E501

        if not sid and not username:
            raise ValueError("Either sid or username argument is required")

        if not sid and username:
            sid = self._users.get(username)

        if not sid:
            raise ValueError("No SID provided or no SID found")

        try:
            blob = DPAPIBlob(data)
        except EOFError as e:
            raise ValueError(f"Failed to parse DPAPI blob: {e}")

        if not (mk := self.master_keys.get(sid, {}).get(blob.guid)):
            raise ValueError(f"Blob is encrypted using master key {blob.guid} that we do not have for SID {sid}")

        if not blob.decrypt(mk.key, **kwargs):
            raise ValueError(f"Failed to decrypt blob for SID {sid}")

        return blob.clear_text

    def decrypt_blob(self, data: bytes, **kwargs) -> bytes:
        """Attempt to decrypt the given bytes using any of the available master keys.

        Args:
            data: Bytes of DPAPI blob to decrypt.
            **kwargs: Arbitrary named arguments to pass to :meth:`DPAPIBlob.decrypt <dissect.target.plugins.os.windows.dpapi.blob.Blob.decrypt>` function.

        Raises:
            ValueError: When conditions to decrypt are not met or if decrypting failed.

        Returns:
            Decrypted bytes.
        """  # noqa: E501
        try:
            blob = DPAPIBlob(data)
        except EOFError as e:
            raise ValueError(f"Failed to parse DPAPI blob: {e}")

        for user in self.master_keys:
            for mk in self.master_keys[user].values():
                if blob.decrypt(mk.key, **kwargs):
                    return blob.clear_text

        raise ValueError("Failed to decrypt blob using any available master key")



from __future__ import annotations

import io
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

from dissect.target.exceptions import ContainerError
from dissect.target.helpers.lazy import import_lazy
from dissect.target.helpers.logging import get_logger
from dissect.target.helpers.utils import readinto

if TYPE_CHECKING:
    from dissect.target.volume import VolumeSystem

CONTAINERS: list[type[Container]] = []
MODULE_PATH = "dissect.target.containers"

RawContainer = import_lazy("dissect.target.containers.raw").RawContainer
"""A lazy import of :mod:`dissect.target.containers.raw`."""

log = get_logger(__name__)


class Container(io.IOBase):
    """Base class that acts as a file-like object wrapper around anything that can behave like a "raw disk".

    Containers are anything from raw disk images and virtual disks, to evidence containers and made-up binary formats.
    Consumers of the ``Container`` class only need to implement ``seek``, ``tell`` and ``read``.
    Override ``__init__`` for any opening that you may need to do, but don't forget to initialize the super class.

    Args:
        fh: The source file-like object of the container or a ``Path`` object to the file.
        size: The size of the container.
        vs: An optional shorthand to set the underlying volume system, usually set later.
    """

    # Due to lazy importing we generally can't use isinstance(), so we add a short identifying string to each class
    # This has the added benefit of having a readily available "pretty name" for each implementation
    __type__: str = None
    """A short string identifying the type of container."""

    def __init__(self, fh: BinaryIO | Path, size: int, vs: VolumeSystem | None = None):
        self.fh = fh
        self.size = size

        # Shorthand access to vs
        self.vs = vs

        if self.__type__ is None:
            raise NotImplementedError(f"{self.__class__.__name__} must define __type__")

    def __repr__(self) -> str:
        return f"<Container type={self.__type__} size={self.size} vs={self.vs.__type__ if self.vs else None}>"

    @classmethod
    def detect(cls, item: list | BinaryIO | Path) -> bool:
        """Detect if this ``Container`` can handle this file format.

        Args:
            item: The object we want to see if it can be handled by this ``Container``.

        Returns:
            ``True`` if this ``Container`` can be used, ``False`` otherwise.
        """
        i = item[0] if isinstance(item, list) else item
        if hasattr(i, "read"):
            return cls.detect_fh(i, item)
        if cls.detect_path(i, item):
            return True
        if i.exists():
            with i.open("rb") as fh:
                return cls.detect_fh(fh, item)

        return False

    @classmethod
    def detect_fh(cls, fh: BinaryIO, original: list | BinaryIO) -> bool:
        """Detect if this ``Container`` can be used to open the file-like object ``fh``.

        The function checks whether the raw data contains any magic information that corresponds to this
        specific container.

        Args:
            fh: A file-like object that we want to open a ``Container`` on.
            original: The original argument passed to ``detect()``.

        Returns:
            ``True`` if this ``Container`` can be used for this file-like object, ``False`` otherwise.
        """
        offset = fh.tell()
        try:
            fh.seek(0)
            return cls._detect_fh(fh, original)
        except NotImplementedError:
            raise
        except Exception as e:
            log.warning("Failed to detect %s container", cls.__name__)
            log.debug("", exc_info=e)
        finally:
            fh.seek(offset)

        return False

    @staticmethod
    def _detect_fh(fh: BinaryIO, original: list | BinaryIO) -> bool:
        """Detect if this ``Container`` can be used to open the file-like object ``fh``.

        This method should be implemented by subclasses. The position of ``fh`` is guaranteed to be ``0``.

        Args:
            fh: A file-like object that we want to open a ``Container`` on.
            original: The original argument passed to ``detect()``.

        Returns:
            ``True`` if this ``Container`` can be used for this file-like object, ``False`` otherwise.
        """
        raise NotImplementedError

    @staticmethod
    def detect_path(path: Path, original: list | Path) -> bool:
        """Detect if this ``Container`` can be used to open ``path``.

        The function checks wether file inside ``path`` is formatted in such a way that
        this ``Container`` can be used to read it. For example, it validates against the
        file extension.

        Args:
            path: A location to a file.
            original: The original argument passed to ``detect()``.

        Returns:
            ``True`` if this ``Container`` can be used for this path, ``False`` otherwise.
        """
        raise NotImplementedError

    def read(self, length: int = -1) -> bytes:
        """Read a ``length`` of bytes from this ``Container``."""
        raise NotImplementedError

    def readinto(self, b: bytearray) -> int:
        """Uses :func:`dissect.target.helpers.utils.readinto`."""
        return readinto(buffer=b, fh=self)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        """Change the stream position to ``offset``.

        ``whence`` determines where to seek from:

        * ``io.SEEK_SET`` (``0``):: absolute offset in the stream.
        * ``io.SEEK_CUR`` (``1``):: current position in the stream.
        * ``io.SEEK_END`` (``2``):: end of stream.

        Args:
            offset: The offset relative to the position indicated by ``whence``.
            whence: Where to start the seek from.
        """
        raise NotImplementedError

    def seekable(self) -> bool:
        """Returns whether ``seek`` can be used by this ``Container``. Always ``True``."""
        return True

    def tell(self) -> int:
        """Returns the current seek position of the ``Container``."""
        raise NotImplementedError

    def close(self) -> None:
        """Close the container.

        Override this if you need to clean-up anything.
        """


def register(module: str, class_name: str, internal: bool = True) -> None:
    """Register a container implementation to use when opening a container.

    This function registers a container using ``module`` relative to the ``MODULE_PATH``.
    It lazily imports the module, and retrieves the specific class from it.

    Args:
        module: The module where to find the container.
        class_name: The class to load.
        internal: Whether it is an internal module or not.
    """

    if internal:
        module = f"{MODULE_PATH}.{module}"

    CONTAINERS.append(getattr(import_lazy(module), class_name))


def open(item: list | str | BinaryIO | Path, *args, **kwargs) -> Container:
    """Open a :class:`Container` from the given object.

    All currently supported containers are checked to find a compatible one.
    :class:`RawContainer <dissect.target.containers.raw.RawContainer>` must always be checked last
    since it always succeeds!

    Args:
        item: The object we want to open a :class`Container` from.

    Raises:
        ContainerError: When a compatible :class`Container` was found but it failed to open.
        ContainerError: When no compatible :class`Container` implementations were found.
    """
    if isinstance(item, list):
        item = [Path(entry) if isinstance(entry, str) else entry for entry in item]
    elif isinstance(item, str):
        item = Path(item)

    first = item[0] if isinstance(item, list) else item
    first_fh = None
    first_fh_opened = False
    first_fh_offset = None
    first_path = None

    if hasattr(first, "read"):
        first_fh = first
    else:
        first_path = first
        if first_path.is_file():
            first_fh = first.open("rb")
            first_fh_opened = True

    if first_fh:
        first_fh_offset = first_fh.tell()
        first_fh.seek(0)

    try:
        for container in [*CONTAINERS, RawContainer]:
            try:
                # Path must be leading for things like SplitContainer, but fall back to fh if we have one
                if (first_path and container.detect_path(first_path, item)) or (
                    first_fh and container.detect_fh(first_fh, item)
                ):
                    return container(item, *args, **kwargs)
            except ImportError as e:  # noqa: PERF203
                log.info("Failed to import %s", container)
                log.debug("", exc_info=e)
            except Exception as e:
                raise ContainerError(f"Failed to open container {item}") from e
    finally:
        if first_fh_opened:
            first_fh.close()
        elif first_fh:
            first_fh.seek(first_fh_offset)

    raise ContainerError(f"Failed to detect container for {item}")


register("ewf", "EwfContainer")
register("vmdk", "VmdkContainer")
register("vhdx", "VhdxContainer")
register("vhd", "VhdContainer")
register("qcow2", "QCow2Container")
register("vdi", "VdiContainer")
register("hdd", "HddContainer")
register("hds", "HdsContainer")
register("split", "SplitContainer")
register("fortifw", "FortiFirmwareContainer")



from __future__ import annotations

import logging
from collections.abc import Callable
from functools import partial
from typing import TYPE_CHECKING

from flow.record import GroupedRecord, Record, RecordDescriptor, fieldtypes

from dissect.target.exceptions import FileNotFoundError, FilesystemError
from dissect.target.helpers.hashutil import common
from dissect.target.helpers.utils import StrEnum
from dissect.target.target import Target

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from dissect.target.helpers.fsutil import TargetPath

__all__ = ("Modifier", "ModifierFunc", "get_modifier_function")

RECORD_NAME = "filesystem/file/digest"
NAME_SUFFIXES = ["_resolved", "_digest"]
RECORD_TYPES = ["path", "digest"]

ModifierFunc = Callable[[Target, Record], GroupedRecord]


class Modifier(StrEnum):
    RESOLVE = "resolve"
    HASH = "hash"


def _create_modified_record(
    record_name: str, field_name: str, field_info: Iterable[tuple[str, str, TargetPath]]
) -> Record:
    record_kwargs = {}
    record_def = []
    for type, name_suffix, data in field_info:
        extended_field_name = f"{field_name}{name_suffix}"
        record_kwargs.update({extended_field_name: data})
        record_def.append((type, extended_field_name))

    _record = RecordDescriptor(record_name, record_def)
    return _record(**record_kwargs)


def _resolve_path_records(field_name: str, resolved_path: TargetPath) -> Record:
    """Resolve files from path fields inside the record."""
    type_info = [("path", "_resolved", resolved_path)]
    return _create_modified_record("filesystem/file/resolved", field_name, type_info)


def _hash_path_records(field_name: str, resolved_path: TargetPath) -> Record:
    """Hash files from path fields inside the record.

    Args:
        field_name: Name of the field.
        resolved_path: Path to the file we should hash.

    Raises:
        FileNotFoundError: Raised if the provided ``resolved_path`` does not exist or is not a file on the target.

    Returns:
        Modified record with digests of path field types.
    """

    if not resolved_path.exists() or not resolved_path.is_file():
        raise FileNotFoundError(f"Path not found or is not a file: '{resolved_path}'")

    with resolved_path.open() as fh:
        path_hash = common(fh)

    type_info = zip(RECORD_TYPES, NAME_SUFFIXES, [resolved_path, path_hash], strict=False)

    return _create_modified_record("filesystem/file/digest", field_name, type_info)


MODIFIER_MAPPING = {
    Modifier.RESOLVE: _resolve_path_records,
    Modifier.HASH: _hash_path_records,
}


def _resolve_path_types(target: Target, record: Record) -> Iterator[tuple[str, TargetPath]]:
    for field_name, field_type in record._field_types.items():
        if not issubclass(field_type, (fieldtypes.path, fieldtypes.command)):
            continue

        path = getattr(record, field_name, None)
        if path is None:
            continue

        if isinstance(path, fieldtypes.command):
            path = path.executable

        yield field_name, target.resolve(str(path))


def modify_record(target: Target, record: Record, modifier_function: ModifierFunc) -> GroupedRecord:
    additional_records = []

    for field_name, resolved_path in _resolve_path_types(target, record):
        try:
            _record = modifier_function(field_name, resolved_path)
        except FilesystemError as e:  # noqa: PERF203
            level = logging.INFO if isinstance(e, FileNotFoundError) else logging.WARNING
            target.log.log(
                level,
                "Unable to modify record '%s' with function '%s': %s",
                record._desc.name,
                modifier_function.__name__,
                e,
            )
            target.log.debug("", exc_info=e)

        else:
            additional_records.append(_record)

    if not additional_records:
        return record

    return GroupedRecord(record._desc.name, [record, *additional_records])


def _noop(_target: Target, record: Record) -> Record:
    return record


def get_modifier_function(modifier_type: Modifier) -> ModifierFunc:
    if func := MODIFIER_MAPPING.get(modifier_type):
        return partial(modify_record, modifier_function=func)

    return _noop



from __future__ import annotations

import textwrap
from datetime import timedelta, timezone
from io import BytesIO
from typing import TYPE_CHECKING

from flow.record.fieldtypes import datetime as dt

from dissect.target.plugins.apps.webserver.caddy import CaddyPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_caddy_txt(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    tz = timezone(timedelta(hours=-7))
    fs_unix.map_file_fh(
        "var/log/caddy_access.log",
        BytesIO(b'127.0.0.1 - - [10/Oct/2000:13:55:36 -0700] "GET /apache_pb.gif HTTP/1.1" 200 2326'),
    )

    target_unix.add_plugin(CaddyPlugin)
    results = list(target_unix.caddy.access())

    assert len(results) == 1

    record = results[0]
    assert record.remote_ip == "127.0.0.1"
    assert record.ts == dt(2000, 10, 10, 13, 55, 36, tzinfo=tz)
    assert record.method == "GET"
    assert record.uri == "/apache_pb.gif"
    assert record.protocol == "HTTP/1.1"
    assert record.status_code == 200
    assert record.bytes_sent == 2326


def test_caddy_json(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file(
        "var/log/caddy_access.log",
        absolute_path("_data/plugins/apps/webserver/caddy/access.log"),
    )

    target_unix.add_plugin(CaddyPlugin)
    results = list(target_unix.caddy.access())

    assert len(results) == 2

    record = results[0]
    assert record.remote_ip == "172.17.0.1"
    assert record.ts == dt(2023, 2, 6, 15, 5, 49, 64393, tzinfo=timezone.utc)
    assert record.method == "GET"
    assert record.uri == "/"
    assert record.protocol == "HTTP/1.1"
    assert record.status_code == 200
    assert record.bytes_sent == 12


def test_caddy_config(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    config_file = absolute_path("_data/plugins/apps/webserver/caddy/Caddyfile")
    fs_unix.map_file("etc/caddy/Caddyfile", config_file)

    fs_unix.map_file("etc/caddy/Caddyfile", config_file)
    fs_unix.map_file_fh("var/www/log/access.log", BytesIO(b"Foo"))
    fs_unix.map_file_fh("var/log/caddy/access.log", BytesIO(b"Foo"))

    log_paths = CaddyPlugin(target_unix).get_log_paths()

    assert len(log_paths) == 2
    assert str(log_paths[0]) == "/var/log/caddy/access.log"
    assert str(log_paths[1]) == "/var/www/log/access.log"


def test_caddy_config_logs_logrotated(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    config_file = absolute_path("_data/plugins/apps/webserver/caddy/Caddyfile")
    fs_unix.map_file("etc/caddy/Caddyfile", config_file)

    fs_unix.map_file_fh("var/www/log/access.log", BytesIO(b"Foo"))
    fs_unix.map_file_fh("var/www/log/access.log.1", BytesIO(b"Foo1"))
    fs_unix.map_file_fh("var/www/log/access.log.2", BytesIO(b"Foo2"))
    fs_unix.map_file_fh("var/www/log/access.log.3", BytesIO(b"Foo3"))

    log_paths = CaddyPlugin(target_unix).get_log_paths()

    assert len(log_paths) == 4


def test_caddy_config_commented(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    config = """
    root /var/www/html
    1.example.com {
        log {
            # output file log/old.log
            output file log/new.log
        }
    }
    # 2.example.com {
    #     log {
    #         output file /completely/disabled/access.log
    #     }
    # }
    """
    fs_unix.map_file_fh("etc/caddy/Caddyfile", BytesIO(textwrap.dedent(config).encode()))
    fs_unix.map_file_fh("var/www/log/old.log", BytesIO(b"Foo"))
    fs_unix.map_file_fh("var/www/log/new.log", BytesIO(b"Foo"))
    fs_unix.map_file_fh("completely/disabled/access.log", BytesIO(b"Foo"))

    log_paths = CaddyPlugin(target_unix).get_log_paths()

    assert len(log_paths) == 3
    assert str(log_paths[0]) == "/var/www/log/old.log"
    assert str(log_paths[1]) == "/var/www/log/new.log"
    assert str(log_paths[2]) == "/completely/disabled/access.log"



from __future__ import annotations

from datetime import datetime
from functools import cached_property
from typing import TYPE_CHECKING, Any

from dissect.database.ese.ntds import NTDS

from dissect.target.exceptions import RegistryKeyNotFoundError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, UnsupportedPluginError, export, internal
from dissect.target.plugins.os.windows.sam import des_decrypt

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.database.ese.ntds.objects import Computer, User

    from dissect.target.target import Target


GENERIC_FIELDS = [
    ("string", "cn"),
    ("string", "upn"),
    ("string", "sam_name"),
    ("string", "sam_type"),
    ("string", "description"),
    ("string", "sid"),
    ("varint", "rid"),
    ("datetime", "password_last_set"),
    ("datetime", "logon_last_failed"),
    ("datetime", "logon_last_success"),
    ("datetime", "account_expires"),
    ("datetime", "creation_time"),
    ("datetime", "last_modified_time"),
    ("boolean", "admin_count"),
    ("boolean", "is_deleted"),
    ("string", "lm"),
    ("string[]", "lm_history"),
    ("string", "nt"),
    ("string[]", "nt_history"),
    ("string", "supplemental_credentials"),
    ("string", "user_account_control"),
    ("string[]", "object_classes"),
    ("string", "distinguished_name"),
    ("string", "object_guid"),
    ("uint32", "primary_group_id"),
    ("string[]", "member_of"),
    ("string[]", "service_principal_name"),
]

# Record descriptor for NTDS user secrets
NtdsUserRecord = TargetRecordDescriptor(
    "windows/ad/user",
    [
        *GENERIC_FIELDS,
        ("string", "info"),
        ("string", "comment"),
        ("string", "telephone_number"),
        ("string", "home_directory"),
    ],
)
NtdsComputerRecord = TargetRecordDescriptor(
    "windows/ad/computer",
    [
        *GENERIC_FIELDS,
        ("string", "dns_hostname"),
        ("string", "operating_system"),
        ("string", "operating_system_version"),
    ],
)

NtdsGPORecord = TargetRecordDescriptor(
    "windows/ad/gpo",
    [
        ("string", "cn"),
        ("string", "distinguished_name"),
        ("string", "object_guid"),
        ("string", "name"),
        ("string", "display_name"),
        ("datetime", "creation_time"),
        ("datetime", "last_modified_time"),
    ],
)

# NTDS Registry consts
NTDS_PARAMETERS_REGISTRY_PATH = "HKLM\\SYSTEM\\CurrentControlSet\\Services\\NTDS\\Parameters"
NTDS_PARAMETERS_DB_VALUE = "DSA Database file"

# Default values
DEFAULT_LM_HASH = "aad3b435b51404eeaad3b435b51404ee"
DEFAULT_NT_HASH = "31d6cfe0d16ae931b73c59d7e0c089c0"


class NtdsPlugin(Plugin):
    """Plugin to parse NTDS.dit Active Directory database and extract user credentials.

    This plugin extracts user password hashes, password history, Kerberos keys, and other authentication data
    from the NTDS.dit database found on Windows Domain Controllers.
    """

    __namespace__ = "ad"

    def __init__(self, target: Target):
        super().__init__(target)
        path = "sysvol/windows/NTDS/ntds.dit"
        if self.target.has_function("registry"):
            try:
                key = self.target.registry.value(NTDS_PARAMETERS_REGISTRY_PATH, NTDS_PARAMETERS_DB_VALUE)
                path = key.value
            except RegistryKeyNotFoundError:
                pass

        self.path = self.target.fs.path(path)

    def check_compatible(self) -> None:
        if not self.target.has_function("lsa"):
            raise UnsupportedPluginError("System Hive is not present or LSA function not available")

        if not self.path.is_file():
            raise UnsupportedPluginError("No NTDS.dit database found on target")

    @cached_property
    @internal
    def ntds(self) -> NTDS:
        ntds = NTDS(self.path.open("rb"))

        if self.target.has_function("lsa"):
            ntds.pek.unlock(self.target.lsa.syskey)

        return ntds

    @export(record=NtdsUserRecord)
    def users(self) -> Iterator[NtdsUserRecord]:
        """Extract all user accounts from the NTDS.dit database."""
        for user in self.ntds.users():
            yield NtdsUserRecord(
                **extract_user_info(user, self.target),
                info=user.get("info"),
                comment=user.get("comment"),
                telephone_number=user.get("telephoneNumber"),
                home_directory=user.get("homeDirectory"),
                _target=self.target,
            )

    @export(record=NtdsComputerRecord)
    def computers(self) -> Iterator[NtdsComputerRecord]:
        """Extract all computer accounts from the NTDS.dit database."""
        for computer in self.ntds.computers():
            yield NtdsComputerRecord(
                **extract_user_info(computer, self.target),
                dns_hostname=computer.get("dNSHostName"),
                operating_system=computer.get("operatingSystem"),
                operating_system_version=computer.get("operatingSystemVersion"),
                _target=self.target,
            )

    @export(record=NtdsGPORecord)
    def group_policies(self) -> Iterator[NtdsGPORecord]:
        """Extract all group policy objects (GPO) NTDS.dit database."""

        for gpo in self.ntds.group_policies():
            yield NtdsGPORecord(
                cn=gpo.cn,
                distinguished_name=gpo.distinguished_name,
                object_guid=gpo.guid,
                name=gpo.name,
                display_name=gpo.display_name,
                creation_time=gpo.when_created,
                last_modified_time=gpo.when_changed,
                _target=self.target,
            )


def extract_user_info(user: User | Computer, target: Target) -> dict[str, Any]:
    """Extract generic information from a User or Computer account."""

    lm_hash = des_decrypt(lm_pwd, user.rid).hex() if (lm_pwd := user.get("dBCSPwd")) else DEFAULT_LM_HASH
    nt_hash = des_decrypt(nt_pwd, user.rid).hex() if (nt_pwd := user.get("unicodePwd")) else DEFAULT_NT_HASH

    # Decrypt password history
    lm_history = [des_decrypt(lm, user.rid).hex() for lm in user.get("lmPwdHistory")]
    nt_history = [des_decrypt(nt, user.rid).hex() for nt in user.get("ntPwdHistory")]

    try:
        member_of = [group.distinguished_name for group in user.groups()]
    except Exception as e:
        member_of = []
        target.log.warning("Failed to extract group membership for user %s: %s", user, e)
        target.log.debug("", exc_info=e)

    # Extract supplemental credentials and yield records
    return {
        "cn": user.cn,
        "upn": user.get("userPrincipalName"),
        "sam_name": user.sam_account_name,
        "sam_type": user.sam_account_type.name,
        "description": user.get("description"),
        "sid": user.sid,
        "rid": user.rid,
        "password_last_set": user.get("pwdLastSet"),
        "logon_last_failed": user.get("badPasswordTime"),
        "logon_last_success": user.get("lastLogon"),
        "account_expires": user.get("accountExpires") if isinstance(user.get("accountExpires"), datetime) else None,
        "creation_time": user.when_created,
        "last_modified_time": user.when_changed,
        "admin_count": user.get("adminCount"),
        "is_deleted": user.is_deleted,
        "lm": lm_hash,
        "lm_history": lm_history,
        "nt": nt_hash,
        "nt_history": nt_history,
        "supplemental_credentials": user.get("supplementalCredentials"),
        "user_account_control": user.user_account_control.name,
        "object_classes": user.object_class,
        "distinguished_name": user.distinguished_name,
        "object_guid": user.guid,
        "primary_group_id": user.primary_group_id,
        "member_of": member_of,
        "service_principal_name": user.get("servicePrincipalName"),
    }



from __future__ import annotations

import codecs
import hashlib
import re
import string
from pathlib import Path
from typing import TYPE_CHECKING

from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from dissect.target.container import Container
    from dissect.target.volume import Volume


re_NOFLAG = 0  # re.NOFLAG is Python 3.11 and newer only

QFindMatchRecord = TargetRecordDescriptor(
    "qfind/match",
    [
        ("string", "disk"),
        ("varint", "offset"),
        ("string", "needle"),
        ("string", "codec"),
        ("bytes", "match"),
        ("bytes", "buffer"),
    ],
)


class QFindPlugin(Plugin):
    """Basically ``grep -a "malware"`` on steroids."""

    def check_compatible(self) -> None:
        pass

    @arg("-n", "--needles", nargs="*", metavar="NEEDLES", help="needles to search for")
    @arg("-nf", "--needle-file", type=Path, help="file containing the needles to search for")
    @arg("-e", "--encoding", help="encode text needles with these comma separated encodings")
    @arg("--regex", action="store_true", help="parse needles as regex patterns")
    @arg("--no-hex-decode", action="store_true", help="do not automatically add decoded hex needles (only in raw mode)")
    @arg("-i", "--ignore-case", action="store_true", help="case insensitive search")
    @arg("-u", "--unique", action="store_true", help="only yield unique string hits (does not apply to raw output)")
    @arg("-W", "--window", type=int, default=256, help="maximum window size in bytes for context around each hit")
    @arg("--strip-null-bytes", action="store_true", help="strip null bytes from matched buffer")
    @export(record=QFindMatchRecord)
    def qfind(
        self,
        needles: list[str] | None = None,
        needle_file: Path | None = None,
        encoding: str = "",
        no_hex_decode: bool = False,
        regex: bool = False,
        ignore_case: bool = False,
        unique: bool = False,
        window: int = 256,
        strip_null_bytes: bool = False,
        *,
        progress: Callable[[Container | Volume, int, int], None] | None = None,
    ) -> Iterator[QFindMatchRecord]:
        """Find a needle in a haystack.

        Hex encode needles starting with ``#`` in needle files, otherwise these needles are ignored.

        Example:
            .. code-block::

                # find all instances of "malware" in the target
                target-qfind <TARGET> --needles malware

                # find all instances of "malware" in the target, ignoring case
                target-qfind <TARGET> --needles MaLwArE --ignore-case

                # find all instances of "malware" in the target and show raw hex dumps
                target-qfind <TARGET> --needles malware --raw

                # find all instances of "malware" in the target, in UTF-8 and UTF-16-LE (UTF-8 is default)
                target-qfind <TARGET> --needles malware --encoding utf-16-le

                # find all matches of regular expression "malware\\s\\d+" in the target (e.g. ``malware 1337``)
                target-qfind <TARGET> --needles "malware\\s\\d+" --regex

                # use a file for needles
                target-qfind <TARGET> --needle-file needles.txt

                # use target-query instead of target-qfind (output in records)
                target-query <TARGET> -f qfind --needles malware
        """
        all_needles = set(needles or [])
        if needle_file and needle_file.exists():
            with needle_file.open("r") as fh:
                for line in fh:
                    if line := line.strip():
                        if line.startswith("#"):
                            self.target.log.warning("Ignoring needle %r", line)
                        else:
                            all_needles.add(line)

        self.target.log.info("Loaded %s needles", len(all_needles))

        encodings = set()
        for codec in (encoding or "").split(","):
            if not (codec := codec.strip()):
                continue

            try:
                codecs.lookup(codec)
            except LookupError:
                self.target.log.warning("Unknown encoding: %s", codec)
            else:
                encodings.add(codec)

        needle_lookup: dict[bytes, tuple[str, str]] = {}
        for needle in all_needles:
            encoded_needle = needle.encode("utf-8")
            needle_lookup[encoded_needle] = (needle, "utf-8")

            if not no_hex_decode and len(needle) % 2 == 0 and all(c in string.hexdigits for c in needle):
                encoded_needle = bytes.fromhex(needle)
                needle_lookup[encoded_needle] = (needle, "hex")

            for codec in encodings:
                try:
                    encoded_needle = needle.encode(codec)
                except UnicodeEncodeError:  # noqa: PERF203
                    self.target.log.warning("Cannot encode needle with %s: %s", codec, needle)
                else:
                    needle_lookup[encoded_needle] = (needle, codec)

        if not needle_lookup:
            self.target.log.error("No needles to search for (use '--needles' or '--needle-file')")
            return

        if ignore_case or regex:
            tmp = {}
            for encoded_needle, _ in needle_lookup.items():
                encoded_needle = encoded_needle if regex else re.escape(encoded_needle)
                tmp[re.compile(encoded_needle, re.IGNORECASE if ignore_case else re_NOFLAG)] = _
            needle_lookup = tmp

        seen = set()
        for disk, stream, needle, offset, match in self.target.scrape.find(
            list(needle_lookup.keys()), progress=progress
        ):
            original_needle, codec = needle_lookup[needle]
            needle_len = len(needle.pattern if isinstance(needle, re.Pattern) else needle)
            before_offset = max(0, offset - window)
            stream.seek(before_offset)
            buf = stream.read((offset - before_offset) + max(window, needle_len))

            if unique:
                digest = hashlib.sha1(buf).digest() if window > 20 else buf

                if digest in seen:
                    continue
                seen.add(digest)

            match = match.group() if match else original_needle.encode()

            yield QFindMatchRecord(
                disk=repr(disk),
                offset=offset,
                needle=original_needle,
                codec=codec,
                match=match,
                buffer=buf.strip(b"\x00") if strip_null_bytes else buf,
                _target=self.target,
            )



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.plugins.os.unix.linux.redhat.yum import YumPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.mark.parametrize(
    "test_file",
    [
        "yum.log",
        "yum.log.1.gz",
        "yum.log.1.bz2",
    ],
)
def test_yum_logs(test_file: str, target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    tz = timezone.utc
    data_file = absolute_path(f"_data/plugins/os/unix/linux/redhat/yum/{test_file}")
    fs_unix.map_file(f"/var/log/{test_file}", data_file)

    # The yum plugin uses a year rollover helper which uses the modification time of a file to determine
    # the starting year
    # A new source checkout would result in different modification timestamps, so mock it to be in 2023
    entry = fs_unix.get(f"/var/log/{test_file}")
    stat_result = entry.stat()
    stat_result.st_mtime = 1704067199

    with patch.object(entry, "stat") as mock_stat:
        mock_stat.return_value = stat_result

        target_unix.add_plugin(YumPlugin)

        results = list(target_unix.yum.logs())
        assert len(results) == 5

        for record in results:
            assert record.package_manager == "yum"

        assert results[0].ts == datetime(2023, 12, 16, 4, 41, 34, tzinfo=tz)
        assert results[0].operation == "install"
        assert results[0].package_name == "unzip-6.0-24.el7_9.x86_64"
        assert results[0].command is None
        assert results[0].requested_by_user is None

        assert results[-1].ts == datetime(2023, 12, 16, 4, 41, 22, tzinfo=tz)
        assert results[-1].operation == "install"
        assert results[-1].package_name == "unzip-6.0-24.el7_9.x86_64"
        assert results[-1].command is None
        assert results[-1].requested_by_user is None



from __future__ import annotations

import json
import re

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.localeutil import normalize_language
from dissect.target.plugin import export
from dissect.target.plugins.os.default.locale import LocalePlugin

RE_CONFIG_TIMEZONE = re.compile(
    r'set ns param.* -timezone "GMT\+(?P<hours>[0-9]+):(?P<minutes>[0-9]+)-.*-(?P<zone_name>.+)"'
)


class CitrixLocalePlugin(LocalePlugin):
    """Citrix Netscaler locale plugin."""

    def check_compatible(self) -> None:
        if self.target.os != "citrix-netscaler":
            raise UnsupportedPluginError("Citrix Netscaler specific plugin loaded on non-Citrix target")

    @export(property=True)
    def timezone(self) -> str | None:
        """Return configured timezone."""
        # Collect timezone from nsconfig/ns.conf or from shell/date.out if exists
        for path in self.target.fs.path("/flash/nsconfig/").glob("ns.conf*"):
            if match := RE_CONFIG_TIMEZONE.search(path.read_text()):
                return match.groupdict()["zone_name"]

        # Netscaler collector specific check:
        # If timezone not set in ns.conf it is often UTC, lets check for that.
        if (path := self.target.fs.path("/shell/date.out")).exists() and "UTC" in path.read_text():
            return "UTC"

        return None

    @export(property=True)
    def language(self) -> list[str] | None:
        """Return configured UI language(s)."""

        found_languages = set()

        # Iterate logon theme languages
        for path in self.target.fs.path("/var/netscaler/logon/themes/").glob("*/*.json"):
            try:
                theme = json.loads(path.read_text())
                if locale := theme.get("locale"):
                    found_languages.add(normalize_language(locale))
            except (json.JSONDecodeError, UnicodeDecodeError) as e:  # noqa: PERF203
                self.target.log.warning("Unable to parse %s: %s", path, e)
                self.target.log.debug("", exc_info=e)

        # Find using regular BSD locations
        found_languages.update(super().language)

        return list(found_languages)



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from dissect.target.helpers.utils import year_rollover_helper
from dissect.target.plugin import export
from dissect.target.plugins.os.unix.history import (
    CommandHistoryPlugin,
    CommandHistoryRecord,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.helpers.record import UnixUserRecord

RE_CITRIX_NETSCALER_BASH_HISTORY_DATE = re.compile(r"(?P<date>[^<]+)\s")

CITRIX_NETSCALER_BASH_HISTORY_RE = re.compile(
    r"""
        (?P<date>[^<]+)
        \s
        <
            (?P<syslog_facility>[^\.]+)
            \.
            (?P<syslog_loglevel>[^>]+)
        >
        \s
        (?P<hostname>[^\s]+)
        \s
        (?P<process_name>[^\[]+)
            \[
                (?P<process_id>\d+)
            \]
        :
        \s
        (?P<username>.*)\s
        on\s
            (?P<destination>[^\s]+)\s
        shell_command=
        \"
            (?P<command>.*)
        \"
    $
    """,
    re.VERBOSE,
)


class CitrixCommandHistoryPlugin(CommandHistoryPlugin):
    """Citrix command history plugin."""

    COMMAND_HISTORY_ABSOLUTE_PATHS = (
        ("citrix-netscaler-bash", "/var/log/bash.log*"),
        ("citrix-netscaler-sh", "/var/log/sh.log*"),
    )
    COMMAND_HISTORY_RELATIVE_PATHS = (
        *CommandHistoryPlugin.COMMAND_HISTORY_RELATIVE_PATHS,
        ("citrix-netscaler-cli", ".nscli_history"),
    )

    def _find_history_files(self) -> list[tuple[str, TargetPath, UnixUserRecord | None]]:
        """Find history files on the target that this plugin can parse."""
        history_files = []
        for shell, history_absolute_path_glob in self.COMMAND_HISTORY_ABSOLUTE_PATHS:
            history_files.extend(
                (shell, path, None) for path in self.target.fs.path("/").glob(history_absolute_path_glob.lstrip("/"))
            )

        # Also utilize the _find_history_files function of the parent class for relative paths.
        history_files.extend(super()._find_history_files())

        return history_files

    def _find_user_by_name(self, username: str) -> UnixUserRecord | None:
        """Cached function to return the matching UnixUserRecord for a given username."""
        if username is None:
            return None

        user_details = self.target.user_details.find(username=username)
        return user_details.user if user_details else None

    @export(record=CommandHistoryRecord)
    def commandhistory(self) -> Iterator[CommandHistoryRecord]:
        """Return shell history for all Citrix users.

        Some entries are returned in reverse chronological order and can contain negative command order integers due
        to the way Citrix stores bash history commands.
        """

        for shell, history_path, user in self._history_files:
            if shell == "citrix-netscaler-cli":
                yield from self.parse_netscaler_cli_history(history_path, user)
            elif shell in ("citrix-netscaler-bash", "citrix-netscaler-sh"):
                yield from self.parse_netscaler_bash_history(history_path)
            else:
                yield from self.parse_generic_history(history_path, user, shell)

    def parse_netscaler_bash_history(self, path: TargetPath) -> Iterator[CommandHistoryRecord]:
        """Parse bash.log* contents."""

        i = 0
        for ts, line in year_rollover_helper(path, RE_CITRIX_NETSCALER_BASH_HISTORY_DATE, "%b %d %H:%M:%S "):
            line = line.strip()
            if not line:
                continue

            match = CITRIX_NETSCALER_BASH_HISTORY_RE.match(line)
            if not match:
                continue

            group = match.groupdict()
            command = group.get("command")
            user = self._find_user_by_name(group.get("username"))

            yield CommandHistoryRecord(
                ts=ts,
                command=command,
                order=-i,  # year_rollover_helper returns entries in reverse order.
                shell="citrix-netscaler-bash",
                source=path,
                _target=self.target,
                _user=user,
            )

            i += 1

    def parse_netscaler_cli_history(
        self, history_file: TargetPath, user: UnixUserRecord
    ) -> Iterator[CommandHistoryRecord]:
        """Parses the history file of the Citrix Netscaler CLI.

        The only difference compared to generic bash history files is that the first line will start with
        ``_HiStOrY_V2_``, which we will skip.
        """
        i = 0
        for line in history_file.open("rt"):
            if not (line := line.strip()):
                continue

            if i == 0 and line == "_HiStOrY_V2_":
                continue

            yield CommandHistoryRecord(
                ts=None,
                command=line,
                order=i,
                shell="citrix-netscaler-cli",
                source=history_file,
                _target=self.target,
                _user=user,
            )

            i += 1



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from functools import lru_cache
from itertools import chain
from typing import TYPE_CHECKING, Any, Final

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import DynamicDescriptor, TargetRecordDescriptor
from dissect.target.helpers.regex.ipaddress import extract_ips
from dissect.target.helpers.utils import year_rollover_helper
from dissect.target.plugin import Plugin, alias, export
from dissect.target.plugins.os.unix.log.helpers import (
    RE_LINE,
    RE_TS,
    is_iso_fmt,
    iso_readlines,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime
    from pathlib import Path

    from dissect.target.target import Target


RE_USER = re.compile(r"for ([^\s]+)")


class BaseService(ABC):
    @classmethod
    @abstractmethod
    def parse(cls, message: str) -> dict[str, any]:
        pass


class SudoService(BaseService):
    """Parsing of sudo service messages in the auth log."""

    RE_SUDO_COMMAND = re.compile(
        r"""
        TTY=(?P<tty>\w+\/\w+)\s;\s          # The TTY -> TTY=pts/0 ;
        PWD=(?P<pwd>[\/\w]+)\s;\s           # The current working directory -> PWD="/home/user" ;
        USER=(?P<effective_user>\w+)\s;\s   # The effective user -> USER=root ;
        COMMAND=(?P<command>.+)$            # The command -> COMMAND=/usr/bin/whoami
        """,
        re.VERBOSE,
    )

    @classmethod
    def parse(cls, message: str) -> dict[str, str]:
        """Parse auth log message from sudo."""
        if not (match := cls.RE_SUDO_COMMAND.search(message)):
            return {}

        return match.groupdict()


class SshdService(BaseService):
    """Class for parsing sshd messages in the auth log."""

    RE_SSHD_PORTREGEX = re.compile(r"port\s(\d+)")
    RE_USER = re.compile(r"for\s([^\s]+)")

    @classmethod
    def parse(cls, message: str) -> dict[str, str | int]:
        """Parse message from sshd."""
        additional_fields = {}
        if ip_address := extract_ips(message):
            field_name = "host_ip" if "listening" in message else "remote_ip"

            if len(ip_address) > 1:
                field_name += "s"

            additional_fields[field_name] = ip_address[0] if len(ip_address) == 1 else ip_address

        if port := cls.RE_SSHD_PORTREGEX.search(message):
            additional_fields["port"] = int(port.group(1))
        if user := cls.RE_USER.search(message):
            additional_fields["user"] = user.group(1)
        # Accepted publickey for test_user from 8.8.8.8 IP port 12345 ssh2: RSA SHA256:123456789asdfghjklertzuio
        if "Accepted publickey" in message:
            ssh_protocol, encryption_algo, key_info = message.split()[-3:]
            hash_algo, key_hash = key_info.split(":")
            additional_fields["ssh_protocol"] = ssh_protocol.strip(":")
            additional_fields["encryption_algorithm"] = encryption_algo
            additional_fields["hash_algorithm"] = hash_algo
            additional_fields["key_hash"] = key_hash
        if (failed := "Failed" in message) or "Accepted" in message:
            action_type = "failed" if failed else "accepted"
            additional_fields["action"] = f"{action_type} authentication"
            additional_fields["authentication_type"] = "password" if "password" in message else "publickey"

        return additional_fields


class SystemdLogindService(BaseService):
    """Class for parsing systemd-logind messages in the auth log.

    References:
        - https://github.com/systemd/systemd/blob/main/src/login/logind-session.c
        - https://github.com/systemd/systemd/blob/main/src/login/logind-session.h
        - https://github.com/systemd/systemd/blob/main/src/login/logind-seat.c
    """

    RE_MSG_WATCHING = re.compile(
        r"""
        (?P<action>Watching\ssystem\sbuttons)\s # Action is "Watching system buttons"
        on\s(?P<device>[^\s]+)\s                # The device the button is related to -> /dev/input/event0
        \((?P<device_name>.*?)\)                # The device (button) name -> (Power button)
        """,
        re.VERBOSE,
    )

    RE_MSG_SESSION_NEW = re.compile(
        r"New session (?P<session>\d+) of user (?P<user>\S+)\.",
    )

    # Reference: https://github.com/systemd/systemd/commit/07b3556510440f7b5f19e19cca1b46d9d581da43
    RE_MSG_SESSION_NEW_VARIANT = re.compile(
        r"New session '(?P<session>\d+)' of user '(?P<user>\S+)' with class '(?P<class>\S+)' and type '(?P<type>\S+)'\."
    )

    RE_MSG_SESSION_LOGGED_OUT = re.compile(
        r"Session (?P<session>\d+) logged out\.",
    )

    RE_MSG_SESSION_REMOVED = re.compile(
        r"Removed session (?P<session>\d+)\.",
    )

    RE_MSG_SEAT = re.compile(r"(?P<action>New seat|Removed seat) (?P<seat>\S+)\.")

    @classmethod
    def parse(cls, message: str) -> dict[str, str]:
        """Parse auth log message from systemd-logind."""
        additional_fields = {}

        # Example: Nov 14 07:14:09 ubuntu-1 systemd-logind[4]: Removed session 4.
        if "Removed session" in message:
            if match := cls.RE_MSG_SESSION_REMOVED.search(message):
                fields = match.groupdict()
                additional_fields["action"] = "removed session"
                additional_fields["session"] = int(fields["session"])

        elif "Watching" in message and (match := cls.RE_MSG_WATCHING.search(message)):
            additional_fields.update(match.groupdict())

        # Example: New session 4 of user sampleuser.
        # Example: New session '4' of user 'sampleuser' with class 'SESSION_USER' and type 'SESSION_TTY'.
        elif "New session" in message:
            if (match := cls.RE_MSG_SESSION_NEW.search(message)) or (
                match := cls.RE_MSG_SESSION_NEW_VARIANT.search(message)
            ):
                fields = match.groupdict()
                additional_fields.update(fields)
                additional_fields["action"] = "new session"
                additional_fields["session"] = int(fields["session"])

        # Example: Session 4 logged out. Waiting for processes to exit.
        elif "logged out" in message:
            if match := cls.RE_MSG_SESSION_LOGGED_OUT.search(message):
                fields = match.groupdict()
                additional_fields["action"] = "logged out session"
                additional_fields["session"] = int(fields["session"])

        # Example: New seat seat0.
        # Example: Removed seat seat0.
        elif ("New seat" in message or "Removed seat" in message) and (match := cls.RE_MSG_SEAT.search(message)):
            fields = match.groupdict()
            additional_fields.update(fields)
            additional_fields["action"] = fields["action"].lower()

        return additional_fields


class SuService(BaseService):
    """Class for parsing su messages in the auth log."""

    RE_SU_BY = re.compile(r"by\s([^\s]+)")
    RE_SU_ON = re.compile(r"on\s([^\s]+)")
    RE_SU_COMMAND = re.compile(r"'(.*?)'")

    @classmethod
    def parse(cls, message: str) -> dict[str, str]:
        additional_fields = {}
        if user := RE_USER.search(message):
            additional_fields["user"] = user.group(1)
        if by := cls.RE_SU_BY.search(message):
            additional_fields["by"] = by.group(1)
        if on := cls.RE_SU_ON.search(message):
            additional_fields["device"] = on.group(1)
        if command := cls.RE_SU_COMMAND.search(message):
            additional_fields["command"] = command.group(1)
        if (failed := "failed" in message) or "Successful" in message:
            additional_fields["su_result"] = "failed" if failed else "success"

        return additional_fields


class PkexecService(BaseService):
    """Class for parsing pkexec messages in the auth log."""

    RE_PKEXEC_COMMAND = re.compile(
        r"""
        (?P<user>\S+?):\sExecuting\scommand\s   # Starts with actual user -> user:
        \[USER=(?P<effective_user>[^\]]+)\]\s   # The impersonated user -> [USER=root]
        \[TTY=(?P<tty>[^\]]+)\]\s               # The tty -> [TTY=unknown]
        \[CWD=(?P<cwd>[^\]]+)\]\s               # Current working directory -> [CWD=/home/user]
        \[COMMAND=(?P<command>[^\]]+)\]         # Command -> [COMMAND=/usr/lib/example]
        """,
        re.VERBOSE,
    )

    @classmethod
    def parse(cls, message: str) -> dict[str, str]:
        """Parse auth log message from pkexec."""
        additional_fields = {}
        if exec_cmd := cls.RE_PKEXEC_COMMAND.search(message):
            additional_fields["action"] = "executing command"
            for key, value in exec_cmd.groupdict().items():
                if value and value.isdigit():
                    value = int(value)
                additional_fields[key] = value

        return additional_fields


class PamUnixService(BaseService):
    RE_PAM_UNIX = re.compile(
        r"""
        pam_unix\([^\s]+:session\):\s(?P<action>session\s\w+)\s     # Session action, usually opened or closed
        for\suser\s(?P<user>[^\s\(]+)(?:\(uid=(?P<user_uid>\d+)\))? # User may contain uid like: root(uid=0)
        (?:\sby\s\(uid=(?P<by_uid>\d+)\))?$                         # Opened action also contains by
        """,
        re.VERBOSE,
    )

    @classmethod
    def parse(cls, message: str) -> dict[str, str | int]:
        """Parse auth log message from pluggable authentication modules (PAM)."""
        if not (match := cls.RE_PAM_UNIX.search(message)):
            return {}

        additional_fields = {}
        for key, value in match.groupdict().items():
            if value and value.isdigit():
                value = int(value)
            additional_fields[key] = value

        return additional_fields


class AuthLogRecordBuilder:
    """Class for dynamically creating auth log records."""

    RECORD_NAME = "linux/log/auth"
    SERVICES: Final[dict[str, BaseService]] = {
        "su": SuService,
        "sudo": SudoService,
        "sshd": SshdService,
        "systemd-logind": SystemdLogindService,
        "pkexec": PkexecService,
    }

    def __init__(self, target: Target):
        self._create_event_descriptor = lru_cache(4096)(self._create_event_descriptor)
        self.target = target

    def _parse_additional_fields(self, service: str | None, message: str) -> dict[str, Any]:
        """Parse additional fields in the message based on the service."""
        if "pam_unix(" in message:
            return PamUnixService.parse(message)

        if service not in self.SERVICES:
            self.target.log.debug("Service %s is not recognised, no additional fields could be parsed", service)
            return {}

        try:
            return self.SERVICES[service].parse(message)
        except Exception as e:
            self.target.log.warning("Parsing additional fields in message '%s' for service %s failed", message, service)
            self.target.log.debug("", exc_info=e)
            raise

    def build_record(self, ts: datetime, source: Path, line: str) -> TargetRecordDescriptor:
        """Builds an ``AuthLog`` event record."""

        record_fields = [
            ("datetime", "ts"),
            ("path", "source"),
            ("string", "service"),
            ("varint", "pid"),
            ("string", "message"),
        ]

        record_values = {
            "ts": ts,
            "message": line,
            "service": None,
            "pid": None,
            "source": source,
            "_target": self.target,
        }

        match = RE_LINE.search(line)
        if match:
            record_values.update(match.groupdict())

        for key, value in self._parse_additional_fields(record_values["service"], line).items():
            record_type = "string"

            if isinstance(value, int):
                record_type = "varint"

            elif key.startswith(("remote_ip", "host_ip")):
                record_type = "net.ipaddress"

            if isinstance(value, list):
                record_type += "[]"

            record_fields.append((record_type, key))
            record_values[key] = value

        # tuple conversion here is needed for lru_cache
        desc = self._create_event_descriptor(tuple(record_fields))
        return desc(**record_values)

    def _create_event_descriptor(self, record_fields: list[tuple[str, str]]) -> TargetRecordDescriptor:
        return TargetRecordDescriptor(self.RECORD_NAME, record_fields)


class AuthPlugin(Plugin):
    """Unix authentication log plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self._auth_log_builder = AuthLogRecordBuilder(target)

    def check_compatible(self) -> None:
        if not any(self._get_paths()):
            raise UnsupportedPluginError("No auth log files found")

    def _get_paths(self) -> Iterator[Path]:
        var_log = self.target.fs.path("/var/log")
        return chain(var_log.glob("auth.log*"), var_log.glob("secure*"))

    @alias("securelog")
    @export(record=DynamicDescriptor(["datetime", "path", "string"]))
    def authlog(self) -> Iterator[Any]:
        """Yield contents of ``/var/log/auth.log*`` and ``/var/log/secure*`` files.

        Order of returned events is not guaranteed to be chronological because of year
        rollover detection efforts for log files without a year in the timestamp.

        The following timestamp formats are recognised automatically. This plugin
        assumes that no custom ``date_format`` template is set in ``syslog-ng`` or ``systemd``
        configuration (defaults to ``M d H:M:S``).

        ISO formatted authlog entries are parsed as can be found in Ubuntu 24.04 and later.

        .. code-block:: text

            CentOS format: Jan 12 13:37:00 hostname daemon: message
            Debian format: Jan 12 13:37:00 hostname daemon[pid]: pam_unix(daemon:session): message
            Ubuntu  24.04: 2024-01-12T13:37:00.000000+02:00 hostname daemon[pid]: pam_unix(daemon:session): message

        References:
            - https://help.ubuntu.com/community/LinuxLogFiles
        """
        target_tz = self.target.datetime.tzinfo

        for auth_file in self.get_paths():
            if is_iso_fmt(auth_file):
                iterable = iso_readlines(auth_file)
            else:
                iterable = year_rollover_helper(auth_file, RE_TS, "%b %d %H:%M:%S", target_tz)

            for ts, line in iterable:
                yield self._auth_log_builder.build_record(ts, auth_file, line)



from __future__ import annotations

import re
from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target

MssqlErrorlogRecord = TargetRecordDescriptor(
    "microsoft/sql/errorlog",
    [
        ("datetime", "ts"),
        ("string", "instance"),
        ("string", "process"),
        ("string", "message"),
        ("path", "path"),
    ],
)

RE_TIMESTAMP_PATTERN = re.compile(r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}.\d{2}")


class MssqlPlugin(Plugin):
    """Return information related to Microsoft SQL Server.

    Currently returns ERRORLOG messages. These log files contain information such as:
        - Logon failures
        - Enabling/disabling of features, such as xp_cmdshell

    References:
        - https://learn.microsoft.com/en-us/sql/relational-databases/logs/view-offline-log-files
    """

    __namespace__ = "mssql"

    MSSQL_KEY_GLOB = "HKLM\\SOFTWARE\\Microsoft\\Microsoft SQL Server\\MSSQL*.*"
    FILE_GLOB = "ERRORLOG*"

    def __init__(self, target: Target):
        super().__init__(target)
        self.instances = self._find_instances()

    def check_compatible(self) -> None:
        if not self.instances:
            raise UnsupportedPluginError("No Microsoft SQL Server instances have been found")

    @export(record=MssqlErrorlogRecord)
    def errorlog(self) -> Iterator[MssqlErrorlogRecord]:
        """Return all Microsoft SQL Server ERRORLOG messages.

        These log files contain information such as:
         - Logon failures
         - Enabling/disabling of features, such as xp_cmdshell

        Yields MssqlErrorlogRecord instances with fields:

        .. code-block:: text

            ts (datetime): Timestamp of the log line.
            instance (str): SQL Server instance name.
            process (str): Process name.
            message (str): Log message.
            path (Path): Path to the log file.

        References:
            - https://learn.microsoft.com/en-us/sql/relational-databases/logs/view-offline-log-files
        """

        for instance, log_path in self.instances:
            for errorlog in log_path.glob(self.FILE_GLOB):
                # The errorlog includes a BOM, so endianess gets determined automatically
                fh = errorlog.open(mode="rt", encoding="utf-16", errors="surrogateescape")

                current_ts: re.Match[str] | None = None
                current_buf = ""

                for line in fh:
                    # If we have a buffer with a timestamp and
                    # our current line also has a timestamp,
                    # we should have a complete record in our buffer.
                    if ts_match := RE_TIMESTAMP_PATTERN.match(line):
                        if current_ts:
                            yield MssqlErrorlogRecord(
                                ts=datetime.strptime(current_ts.group(), "%Y-%m-%d %H:%M:%S.%f").replace(
                                    tzinfo=timezone.utc
                                ),
                                instance=instance,
                                # The process name is a fixed-width field and is always 12 characters long.
                                process=current_buf[23:35].strip(),
                                message=current_buf[35:].strip(),
                                path=errorlog,
                                _target=self.target,
                            )

                        current_ts = ts_match
                        current_buf = line
                    elif current_buf:
                        current_buf += line
                # For the last line
                if current_ts and current_buf:
                    yield MssqlErrorlogRecord(
                        ts=datetime.strptime(current_ts.group(), "%Y-%m-%d %H:%M:%S.%f").replace(tzinfo=timezone.utc),
                        instance=instance,
                        # The process name is a fixed-width field and is always 12 characters long.
                        process=current_buf[23:35].strip(),
                        message=current_buf[35:].strip(),
                        path=errorlog,
                        _target=self.target,
                    )

    def _find_instances(self) -> set[tuple[str, TargetPath]]:
        return {
            (subkey.name, self.target.fs.path(subkey.subkey("SQLServerAgent").value("ErrorLogFile").value).parent)
            for subkey in self.target.registry.glob_ext(self.MSSQL_KEY_GLOB)
        }



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.ntfs import ntfs

from dissect.target.exceptions import RegistryValueNotFoundError, UnsupportedPluginError
from dissect.target.helpers import regutil
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

SyscacheRecord = TargetRecordDescriptor(
    "windows/syscache/object",
    [
        ("datetime", "regf_mtime"),
        ("digest", "digest"),
        ("string", "program_id"),
        ("string", "file_id"),
        ("varint", "object_id"),
        ("varint", "object_lru"),
        ("varint", "usn_journal_id"),
        ("varint", "usn"),
        ("path", "path"),
    ],
)


class SyscachePlugin(Plugin):
    """Plugin to parse Syscache.hve."""

    def __init__(self, target: Target):
        super().__init__(target)
        self.hive = regutil.HiveCollection()

        fpath = self.target.fs.path("sysvol/System Volume Information/Syscache.hve")
        if fpath.exists():
            self.hive.add(regutil.RegfHive(fpath))

    def check_compatible(self) -> None:
        if not len(self.hive) > 0:
            raise UnsupportedPluginError("Could not load Syscache.hve")

    @export(record=SyscacheRecord)
    def syscache(self) -> Iterator[SyscacheRecord]:
        """Parse the objects in the ObjectTable from the Syscache.hve file.

        References:
            - https://dfir.ru/2018/12/02/the-cit-database-and-the-syscache-hive/
        """

        # Try to get the system volume
        mft = None
        sysvol = self.target.fs.mounts["sysvol"]
        if sysvol.__type__ == "ntfs" or hasattr(sysvol, "ntfs"):  # Nasty TarLoader hack
            mft = sysvol.ntfs.mft

        # There's some other stuff here like an IndexTable and LruList
        # Don't think they're too significant, so just iterate over all objects.
        for key in self.hive.keys("DefaultObjectStore\\ObjectTable"):
            for subkey in key.subkeys():
                try:
                    ae_file_id = subkey.value("AeFileID").value.decode("utf-16-le").strip("\x00")
                except RegistryValueNotFoundError:
                    ae_file_id = None

                try:
                    ae_program_id = subkey.value("AeProgramID").value.decode("utf-16-le").strip("\x00")
                except RegistryValueNotFoundError:
                    ae_program_id = None

                try:
                    file_id = subkey.value("_FileId_").value
                except RegistryValueNotFoundError:
                    # Bail out
                    continue

                file_segment = file_id & ((1 << 48) - 1)

                full_path = None
                if mft:
                    try:
                        if path := mft(file_segment).full_path():
                            full_path = self.target.fs.path(f"sysvol\\{path}")
                    except ntfs.Error:
                        pass

                yield SyscacheRecord(
                    regf_mtime=subkey.ts,
                    digest=(None, ae_file_id[4:] if ae_file_id else None, None),
                    program_id=ae_program_id,
                    file_id=f"{file_segment}#{file_id >> 48}",
                    object_id=subkey.value("_ObjectId_").value,
                    object_lru=subkey.value("_ObjectLru_").value,
                    usn_journal_id=subkey.value("_UsnJournalId_").value,
                    usn=subkey.value("_Usn_").value,
                    path=full_path,
                    _target=self.target,
                )



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest

from dissect.target.filesystems.ffs import FfsFilesystem, FfsFilesystemEntry

if TYPE_CHECKING:
    from collections.abc import Iterator

NANOSECONDS_IN_SECOND = 1_000_000_000


@pytest.fixture
def ffs_fs() -> Iterator[FfsFilesystem]:
    with patch("dissect.ffs.ffs.FFS"):
        ffs_fs = FfsFilesystem(Mock())
        ffs_fs.ffs.block_size = 32 * 1024
        yield ffs_fs


@pytest.fixture
def ffs_fs_entry(ffs_fs: FfsFilesystem) -> FfsFilesystemEntry:
    atime = datetime(2024, 10, 1, 12, 0, 0, tzinfo=timezone.utc)
    mtime = datetime(2024, 10, 2, 12, 0, 0, tzinfo=timezone.utc)
    ctime = datetime(2024, 10, 3, 12, 0, 0, tzinfo=timezone.utc)
    btime = datetime(2024, 10, 4, 12, 0, 0, tzinfo=timezone.utc)

    raw_inode = Mock(di_uid=1000, di_nlink=1, di_guid=999, di_size=165002)
    inode = Mock(
        mode=0o100664,
        inum=4,
        inode=raw_inode,
        nblocks=323,
        atime=atime,
        atime_ns=atime.timestamp() * NANOSECONDS_IN_SECOND,
        mtime=mtime,
        mtime_ns=mtime.timestamp() * NANOSECONDS_IN_SECOND,
        ctime=ctime,
        ctime_ns=ctime.timestamp() * NANOSECONDS_IN_SECOND,
        btime=btime,
        btime_ns=btime.timestamp() * NANOSECONDS_IN_SECOND,
        is_file=lambda: True,
        is_dir=lambda: False,
        is_symlink=lambda: False,
    )

    return FfsFilesystemEntry(ffs_fs, "/some_file", inode)


def test_jffs2_stat(ffs_fs_entry: FfsFilesystemEntry) -> None:
    stat = ffs_fs_entry.stat()

    entry = ffs_fs_entry.entry
    assert stat.st_mode == entry.mode
    assert stat.st_ino == entry.inum
    assert stat.st_dev == id(ffs_fs_entry.fs)
    assert stat.st_nlink == entry.inode.di_nlink
    assert stat.st_uid == entry.inode.di_uid
    assert stat.st_gid == entry.inode.di_gid
    assert stat.st_size == entry.inode.di_size
    assert stat.st_atime == entry.atime.timestamp()
    assert stat.st_mtime == entry.mtime.timestamp()
    assert stat.st_ctime == entry.ctime.timestamp()
    assert stat.st_birthtime == entry.btime.timestamp()
    assert stat.st_birthtime_ns == entry.btime_ns
    assert stat.st_blksize == 32 * 1024
    assert stat.st_blocks == 323
